# Copyright 2025 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.
#
# A portion of this file was funded by Blue Origin Enterprises, L.P.
# See https://github.com/OpenC3/cosmos/pull/1963


import os
from contextlib import contextmanager
from openc3.api import WHITELIST
from openc3.api.interface_api import get_interface
from openc3.top_level import DisabledError
from openc3.environment import OPENC3_SCOPE
from openc3.utilities.authorization import authorize
from openc3.utilities.string import simple_formatted
from openc3.models.target_model import TargetModel
from openc3.utilities.extract import *
from openc3.topics.topic import Topic
from openc3.topics.command_topic import CommandTopic
from openc3.topics.interface_topic import InterfaceTopic
from openc3.topics.decom_interface_topic import DecomInterfaceTopic
from openc3.topics.command_decom_topic import CommandDecomTopic
from openc3.packets.packet import Packet

WHITELIST.extend(
    [
        "cmd",
        "cmd_no_range_check",
        "cmd_no_hazardous_check",
        "cmd_no_checks",
        "cmd_raw",
        "cmd_raw_no_range_check",
        "cmd_raw_no_hazardous_check",
        "cmd_raw_no_checks",
        "build_cmd",
        "build_command",  # DEPRECATED
        "enable_cmd",
        "disable_cmd",
        "send_raw",
        "get_all_cmds",
        "get_all_commands",  # DEPRECATED
        "get_all_cmd_names",
        "get_all_command_names",  # DEPRECATED
        "get_cmd",
        "get_command",  # DEPRECATED
        "get_param",
        "get_parameter",  # DEPRECATED
        "get_cmd_buffer",
        "get_cmd_hazardous",
        "get_cmd_value",
        "get_cmd_time",
        "get_cmd_cnt",
        "get_cmd_cnts",
    ]
)


# The following methods sends a command packet to a target. The 'raw' version of the equivalent
# command methods do not perform command parameter conversions.
#
# Accepts two different calling styles:
#   cmd("TGT CMD with PARAM1 val, PARAM2 val")
#   cmd('TGT','CMD',{'PARAM1':val,'PARAM2':val})
#
# Favor the first syntax where possible as it is more succinct.
def cmd(*args, **kwargs):
    return _cmd_implementation("cmd", *args, range_check=True, hazardous_check=True, raw=False, manual=False, **kwargs)


def cmd_raw(*args, **kwargs):
    return _cmd_implementation(
        "cmd_raw", *args, range_check=True, hazardous_check=True, raw=True, manual=False, **kwargs
    )


# S a command packet to a target without performing any value range
# checks on the parameters. Useful for testing to allow sing command
# parameters outside the allowable range as defined in the configuration.
def cmd_no_range_check(*args, **kwargs):
    return _cmd_implementation(
        "cmd_no_range_check",
        *args,
        range_check=False,
        hazardous_check=True,
        raw=False,
        manual=False,
        **kwargs,
    )


def cmd_raw_no_range_check(*args, **kwargs):
    return _cmd_implementation(
        "cmd_raw_no_range_check",
        *args,
        range_check=False,
        hazardous_check=True,
        raw=True,
        manual=False,
        **kwargs,
    )


# S a command packet to a target without performing any hazardous checks
# both on the command itself and its parameters. Useful in scripts to
# prevent popping up warnings to the user.
def cmd_no_hazardous_check(*args, **kwargs):
    return _cmd_implementation(
        "cmd_no_hazardous_check",
        *args,
        range_check=True,
        hazardous_check=False,
        raw=False,
        manual=False,
        **kwargs,
    )


def cmd_raw_no_hazardous_check(*args, **kwargs):
    return _cmd_implementation(
        "cmd_raw_no_hazardous_check",
        *args,
        range_check=True,
        hazardous_check=False,
        raw=True,
        manual=False,
        **kwargs,
    )


# S a command packet to a target without performing any value range
# checks or hazardous checks both on the command itself and its parameters.
def cmd_no_checks(*args, **kwargs):
    return _cmd_implementation(
        "cmd_no_checks",
        *args,
        range_check=False,
        hazardous_check=False,
        raw=False,
        manual=False,
        **kwargs,
    )


def cmd_raw_no_checks(*args, **kwargs):
    return _cmd_implementation(
        "cmd_raw_no_checks",
        *args,
        range_check=False,
        hazardous_check=False,
        raw=True,
        manual=False,
        **kwargs,
    )


# Build a command binary
def build_cmd(*args, range_check=True, raw=False, timeout=5, scope=OPENC3_SCOPE, manual=False):
    match len(args):
        case 1:
            target_name, cmd_name, cmd_params = extract_fields_from_cmd_text(args[0])
        case 2 | 3:
            target_name = args[0]
            cmd_name = args[1]
            if len(args) == 2:
                cmd_params = {}
            else:
                cmd_params = args[2]
        case _:
            # Invalid number of arguments
            raise RuntimeError(f"ERROR: Invalid number of arguments ({len(args)}) passed to build_command()")
    target_name = target_name.upper()
    cmd_name = cmd_name.upper()
    cmd_params = {k.upper(): v for k, v in cmd_params.items()}
    authorize(permission="cmd_info", target_name=target_name, scope=scope, manual=manual)
    return DecomInterfaceTopic.build_cmd(target_name, cmd_name, cmd_params, range_check, raw, timeout, scope)


# build_command is DEPRECATED
build_command = build_cmd


# Helper method for disable_cmd / enable_cmd
@contextmanager
def _get_and_set_cmd(method, *args, scope=OPENC3_SCOPE, manual=False):
    target_name, command_name = _extract_target_command_names(method, *args)
    authorize(
        permission="admin",
        target_name=target_name,
        packet_name=command_name,
        scope=scope,
        manual=manual,
    )
    command = TargetModel.packet(target_name, command_name, type="CMD", scope=scope)
    yield command
    TargetModel.set_packet(target_name, command_name, command, type="CMD", scope=scope)


# @since 5.15.1
def enable_cmd(*args, scope=OPENC3_SCOPE, manual=False):
    with _get_and_set_cmd("enable_cmd", *args, scope=scope, manual=manual) as command:
        command.pop("disabled", None)


# @since 5.15.1
def disable_cmd(*args, scope=OPENC3_SCOPE, manual=False):
    with _get_and_set_cmd("disable_cmd", *args, scope=scope, manual=manual) as command:
        command["disabled"] = True


# Send a raw binary string to the specified interface.
#
# @param interface_name [String] The interface to s the raw binary
# @param data [String] The raw binary data
def send_raw(interface_name, data, scope=OPENC3_SCOPE, manual=False):
    interface_name = interface_name.upper()
    authorize(permission="cmd_raw", interface_name=interface_name, scope=scope, manual=manual)
    get_interface(interface_name, scope=scope)  # Check to make sure the interface exists
    InterfaceTopic.write_raw(interface_name, data, scope=scope)


# Returns the raw buffer from the most recent specified command packet.
#
# @param target_name [String] Target name of the command
# @param command_name [String] Packet name of the command
# @return [Hash] command hash with last command buffer
def get_cmd_buffer(*args, scope=OPENC3_SCOPE, manual=False):
    target_name, command_name = _extract_target_command_names("get_cmd_buffer", *args)
    authorize(
        permission="cmd_info",
        target_name=target_name,
        packet_name=command_name,
        scope=scope,
        manual=manual,
    )
    TargetModel.packet(target_name, command_name, type="CMD", scope=scope)
    topic = f"{scope}__COMMAND__{{{target_name}}}__{command_name}"
    msg_id, message = Topic.get_newest_message(topic)
    if msg_id:
        # Decode the keys for user convenience
        return {k.decode(): v for (k, v) in message.items()}
    return None


# Returns an array of all the commands as a hash
# @param target_name [String] Name of the target
# @return [Array<Hash>] Array of all commands as a hash
def get_all_cmds(target_name, scope=OPENC3_SCOPE, manual=False):
    target_name = target_name.upper()
    authorize(permission="cmd_info", target_name=target_name, scope=scope, manual=manual)
    return TargetModel.packets(target_name, type="CMD", scope=scope)


# get_all_commands is DEPRECATED
get_all_commands = get_all_cmds


# Returns an array of all the command packet names
# @param target_name [String] Name of the target
# @return [Array<String>] Array of all command packet names
def get_all_cmd_names(target_name, hidden=False, scope=OPENC3_SCOPE, manual=False):
    try:
        packets = get_all_cmds(target_name, scope=scope, manual=manual)
    except RuntimeError:
        packets = []
    names = []
    for packet in packets:
        if hidden:
            names.append(packet["packet_name"])
        else:
            if "hidden" not in packet:
                names.append(packet["packet_name"])
    return names


# get_all_command_names is DEPRECATED
get_all_command_names = get_all_cmd_names


# Returns a hash of the given command
def get_cmd(*args, scope=OPENC3_SCOPE, manual=False):
    target_name, command_name = _extract_target_command_names("get_cmd", *args)
    authorize(permission="cmd_info", target_name=target_name, scope=scope, manual=manual)
    return TargetModel.packet(target_name, command_name, type="CMD", scope=scope)


# get_command is DEPRECATED
get_command = get_cmd


# Returns a hash of the given command parameter
# @param target_name [String] Name of the target
# @param command_name [String] Name of the packet
# @param parameter_name [String] Name of the parameter
# @return [Hash] Command parameter as a hash
def get_param(*args, scope=OPENC3_SCOPE, manual=False):
    target_name, command_name, parameter_name = _extract_target_command_parameter_names("get_param", *args)
    authorize(
        permission="cmd_info",
        target_name=target_name,
        packet_name=command_name,
        scope=scope,
        manual=manual,
    )
    return TargetModel.packet_item(target_name, command_name, parameter_name, type="CMD", scope=scope)


# get_parameter is DEPRECATED
get_parameter = get_param


# Returns whether the specified command is hazardous
#
# Accepts two different calling styles:
#   get_cmd_hazardous("TGT CMD with PARAM1 val, PARAM2 val")
#   get_cmd_hazardous('TGT','CMD',{'PARAM1'=>val,'PARAM2'=>val})
#
# @param args [String|Array<String>] See the description for calling style
# @return [Boolean] Whether the command is hazardous
def get_cmd_hazardous(*args, scope=OPENC3_SCOPE, manual=False):
    match len(args):
        case 1:
            target_name, command_name, parameters = extract_fields_from_cmd_text(args[0])
            target_name = target_name.upper()
            command_name = command_name.upper()
            parameters = {k.upper(): v for k, v in parameters.items()}
        case 2 | 3:
            target_name = args[0].upper()
            command_name = args[1].upper()
            if len(args) == 2:
                parameters = {}
            else:
                parameters = args[2]

        case _:
            # Invalid number of arguments
            raise RuntimeError(f"ERROR: Invalid number of arguments ({len(args)}) passed to get_cmd_hazardous()")

    authorize(
        permission="cmd_info",
        target_name=target_name,
        packet_name=command_name,
        scope=scope,
        manual=manual,
    )
    packet = TargetModel.packet(target_name, command_name, type="CMD", scope=scope)
    if packet.get("hazardous") is not None:
        return True

    for item in packet["items"]:
        if item["name"] not in parameters and "states" not in item:
            continue

        # States are an array of the name followed by a dict of 'value' and sometimes 'hazardous'
        for name, items in item["states"].items():
            parameter_name = parameters[item["name"]]
            # Remove quotes from string parameters
            if isinstance(parameter_name, str):
                parameter_name = parameter_name.replace('"', "").replace("'", "")
            # To be hazardous the state must be marked hazardous
            # Check if either the state name or value matches the param passed
            if items.get("hazardous") is not None and (name == parameter_name or items["value"] == parameter_name):
                return True
    return False


# Returns a value from the specified command
def get_cmd_value(
    *args,
    type="CONVERTED",
    scope=OPENC3_SCOPE,
    manual=False,
):
    target_name = None
    command_name = None
    parameter_name = None
    match len(args):
        case 1:
            try:
                target_name, command_name, parameter_name = args[0].upper().split()
            except ValueError:
                # Do nothing because we catch it below
                pass
        case 3:
            target_name = args[0].upper()
            command_name = args[1].upper()
            parameter_name = args[2].upper()
        case 4:
            target_name = args[0].upper()
            command_name = args[1].upper()
            parameter_name = args[2].upper()
            type = args[3].upper()
        case _:
            # Invalid number of arguments
            raise RuntimeError(f"ERROR: Invalid number of arguments ({len(args)}) passed to get_cmd_value()")
    if target_name is None or command_name is None:
        raise RuntimeError(
            f'ERROR: Target name, command name and parameter name required. Usage: get_cmd_value("TGT CMD PARAM") or {method_name}("TGT", "CMD", "PARAM")'
        )

    authorize(
        permission="cmd_info",
        target_name=target_name,
        packet_name=command_name,
        scope=scope,
        manual=manual,
    )
    return CommandDecomTopic.get_cmd_item(target_name, command_name, parameter_name, type=type, scope=scope)


# Returns the time the most recent command was sent
#
# @param target_name [String] Target name of the command. If not given then
#    the most recent time from all commands will be returned
# @param command_name [String] Packet name of the command. If not given then
#    then most recent time from the given target will be returned.
# @return [Array<Target Name, Command Name, Time Seconds, Time Microseconds>]
def get_cmd_time(target_name=None, command_name=None, scope=OPENC3_SCOPE, manual=False):
    authorize(
        permission="cmd_info",
        target_name=target_name,
        packet_name=command_name,
        scope=scope,
        manual=manual,
    )
    if target_name and command_name:
        target_name = target_name.upper()
        command_name = command_name.upper()
        time = CommandDecomTopic.get_cmd_item(
            target_name,
            command_name,
            "RECEIVED_TIMESECONDS",
            type="CONVERTED",
            scope=scope,
        )
        if time is None:
            time = 0
        return (
            target_name,
            command_name,
            int(time),
            int((time - int(time)) * 1_000_000),
        )
    else:
        if not target_name:
            targets = TargetModel.names(scope=scope)
        else:
            targets = [target_name.upper()]

        time = 0
        command_name = None
        for target_name in targets:
            for packet in TargetModel.packets(target_name, type="CMD", scope=scope):
                cur_time = CommandDecomTopic.get_cmd_item(
                    target_name,
                    packet["packet_name"],
                    "RECEIVED_TIMESECONDS",
                    type="CONVERTED",
                    scope=scope,
                )
                if not cur_time:
                    continue

                if cur_time > time:
                    time = cur_time
                    command_name = packet["packet_name"]

        if not command_name:
            target_name = None
        return (
            target_name,
            command_name,
            int(time),
            int((time - int(time)) * 1_000_000),
        )


# Get the transmit count for a command packet
#
# @param target_name [String] Target name of the command
# @param command_name [String] Packet name of the command
# @return [Numeric] Transmit count for the command
def get_cmd_cnt(*args, scope=OPENC3_SCOPE, manual=False):
    target_name, command_name = _extract_target_command_names("get_cmd_cnt", *args)
    authorize(
        permission="system",
        target_name=target_name,
        packet_name=command_name,
        scope=scope,
        manual=manual,
    )
    TargetModel.packet(target_name, command_name, type="CMD", scope=scope)
    return TargetModel.get_command_count(target_name, command_name, scope=scope)


# Get the transmit counts for command packets
#
# @param target_commands [Array<Array<String, String>>] Array of arrays containing target_name, packet_name
# @return [Numeric] Transmit count for the command
def get_cmd_cnts(target_commands, scope=OPENC3_SCOPE):
    authorize(permission="system", scope=scope)
    if isinstance(target_commands, list) and isinstance(target_commands[0], list):
        return TargetModel.get_command_counts(target_commands, scope=scope)
    else:
        raise RuntimeError(
            "get_cmd_cnts takes a dict of dicts containing target, packet_name, e.g. [['INST', 'COLLECT'], ['INST', 'ABORT']]"
        )


def _extract_target_command_names(method_name, *args):
    target_name = None
    command_name = None
    match len(args):
        case 1:
            try:
                target_name, command_name = args[0].upper().split()
            except ValueError:
                # Do nothing because we catch it below
                pass
        case 2:
            target_name = args[0].upper()
            command_name = args[1].upper()
        case _:
            # Invalid number of arguments
            raise RuntimeError(f"ERROR: Invalid number of arguments ({len(args)}) passed to {method_name}()")
    if target_name is None or command_name is None:
        raise RuntimeError(
            f'ERROR: Target name and command name required. Usage: {method_name}("TGT CMD") or {method_name}("TGT", "CMD")'
        )
    return (target_name, command_name)


def _extract_target_command_parameter_names(method_name, *args):
    target_name = None
    command_name = None
    parameter_name = None
    match len(args):
        case 1:
            try:
                target_name, command_name, parameter_name = args[0].upper().split()
            except ValueError:
                # Do nothing because we catch it below
                pass
        case 3:
            target_name = args[0].upper()
            command_name = args[1].upper()
            parameter_name = args[2].upper()
        case _:
            # Invalid number of arguments
            raise RuntimeError(f"ERROR: Invalid number of arguments ({len(args)}) passed to {method_name}()")
    if target_name is None or command_name is None:
        raise RuntimeError(
            f'ERROR: Target name, command name and parameter name required. Usage: {method_name}("TGT CMD PARAM") or {method_name}("TGT", "CMD", "PARAM")'
        )
    return (target_name, command_name, parameter_name)


def _cmd_implementation(
    method_name,
    *args,
    range_check,
    hazardous_check,
    raw,
    manual=False,
    **kwargs,
):
    scope = OPENC3_SCOPE
    if kwargs.get("scope"):
        scope = kwargs["scope"]

    match len(args):
        case 1:
            target_name, cmd_name, cmd_params = extract_fields_from_cmd_text(args[0])
        case 2 | 3:
            target_name = args[0]
            cmd_name = args[1]
            if len(args) == 2:
                cmd_params = {}
            else:
                cmd_params = args[2]
        case _:
            # Invalid number of arguments
            raise RuntimeError(f"ERROR: Invalid number of arguments ({len(args)}) passed to {method_name}()")

    target_name = target_name.upper()
    cmd_name = cmd_name.upper()
    cmd_params = {k.upper(): v for k, v in cmd_params.items()}
    user = authorize(permission="cmd", target_name=target_name, packet_name=cmd_name, scope=scope, manual=manual)
    if not user:
        user = {}
        user["username"] = os.environ.get("OPENC3_MICROSERVICE_NAME")

        # Get the caller stack trace to determine the point in the code where the command was called
        # This code works but ultimately we didn't want to overload 'username' and take a performance hit
        # stack_trace = traceback.extract_stack()
        # for frame in stack_trace:
        #     # Look for the following line in the stack trace which indicates custom code
        #     # File "/tmp/tmpp96e6j83/targets/INST2/lib/example_limits_response.py", line 25, in call"
        #     if f"/targets/{target_name}" in frame.filename:
        #         user = {}
        #         # username is the name of the custom code file
        #         user["username"] = frame.filename.split("/targets/")[-1].split('"')[0]
        #         break

    packet = TargetModel.packet(target_name, cmd_name, type="CMD", scope=scope)
    if packet.get("disabled", False):
        error = DisabledError()
        error.target_name = target_name
        error.cmd_name = cmd_name
        raise error

    timeout = None
    if kwargs.get("timeout") is not None:
        try:
            timeout = float(kwargs["timeout"])
        except ValueError:
            raise RuntimeError(f"Invalid timeout parameter: {timeout}. Must be numeric.")

    # Determine if we should log this command
    log_message = True  # Default is True
    # If the packet has the DISABLE_MESSAGES keyword then no messages
    if packet.get("messages_disabled"):
        log_message = False
    else:
        # Check if any of the parameters have DISABLE_MESSAGES
        for key, value in cmd_params.items():
            found = None
            for item in packet["items"]:
                if item["name"] == key:
                    found = item
                    break
            if (
                found
                and "states" in found
                and value in found["states"]
                and found["states"][value].get("messages_disabled")
            ):
                log_message = False
    # If they explicitly set the log_message kwarg then that overrides the above
    if kwargs.get("log_message") is not None:
        if kwargs["log_message"] not in [True, False]:
            raise RuntimeError(f"Invalid log_message parameter: {kwargs['log_message']}. Must be True or False.")
        log_message = kwargs["log_message"]
    cmd_string = _cmd_log_string(method_name, target_name, cmd_name, cmd_params, packet)

    # Check for the validate kwarg
    validate = True
    if kwargs.get("validate") is not None:
        if kwargs["validate"] not in [True, False]:
            raise RuntimeError(f"Invalid validate parameter: {kwargs['validate']}. Must be True or False.")
        validate = kwargs["validate"]

    username = user["username"] if user and user["username"] else "anonymous"
    command = {
        "target_name": target_name,
        "cmd_name": cmd_name,
        "cmd_params": cmd_params,
        "range_check": str(range_check),
        "hazardous_check": str(hazardous_check),
        "raw": str(raw),
        "cmd_string": cmd_string,
        "username": username,
        "validate": str(validate),
        "manual": str(manual),
        "log_message": str(log_message),
    }
    return CommandTopic.send_command(command, timeout, scope)


def _cmd_log_string(method_name, target_name, cmd_name, cmd_params, packet):
    output_string = f'{method_name}("'
    output_string += target_name + " " + cmd_name
    if not cmd_params:
        output_string += '")'
    else:
        params = []
        for key, value in cmd_params.items():
            if key in Packet.RESERVED_ITEM_NAMES:
                continue

            found = False
            for item in packet["items"]:
                if item["name"] == key:
                    found = item
                    break
            if found and "data_type" in found:
                item_type = found["data_type"]
            else:
                item_type = None

            if isinstance(value, str):
                if item_type == "BLOCK" or item_type == "STRING":
                    if not value.isascii():
                        value = "0x" + simple_formatted(value)
                    else:
                        value = f"'{str(value)}'"
                else:
                    value = convert_to_value(value)
                if len(value) > 256:
                    value = value[:256] + "...'"
                value = value.replace('"', "'")
            elif isinstance(value, list):
                value = f"[{', '.join(str(i) for i in value)}]"
            params.append(f"{key} {value}")
        output_string += " with " + ", ".join(params) + '")'
    return output_string



# Copyright 2023 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

from openc3.script import DISCONNECT
from openc3.script import API_SERVER as API_SERVER

LIMITS_METHODS = [
    "enable_limits",
    "disable_limits",
    "set_limits",
    "enable_limits_group",
    "disable_limits_group",
    "set_limits_set",
]

# Define all the modification methods such that we can disconnect them
for method in LIMITS_METHODS:
    code = [f"def {method}(*args, **kwargs):"]
    if DISCONNECT:
        code.append(f"        print('DISCONNECT: {method}(args) ignored')")
    else:
        code.append(f"        return getattr(API_SERVER, '{method}')(*args, **kwargs)")
    function = compile("\n".join(code), "<string>", "exec")
    exec(function, globals())



# Copyright 2024 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

import json
from openc3.utilities.extract import *
import openc3.script
from openc3.environment import OPENC3_SCOPE


def critical_cmd_status(uuid: str, scope: str = OPENC3_SCOPE):
    try:
        endpoint = f"/openc3-api/criticalcmd/status/{uuid}"
        response = openc3.script.API_SERVER.request("get", endpoint, scope=scope)
        if not response or response.status_code != 200:
            raise RuntimeError(f"Unexpected response to critical_cmd_status: {repr(response)}")
        result = json.loads(response.text)
        return result["status"]
    except Exception as error:
        raise RuntimeError(f"critical_cmd_status failed due to {repr(error)}") from error


def critical_cmd_approve(uuid: str, scope: str = OPENC3_SCOPE):
    try:
        endpoint = f"/openc3-api/criticalcmd/approve/{uuid}"
        response = openc3.script.API_SERVER.request("post", endpoint, json=True, scope=scope)
        if not response or response.status_code != 200:
            if response:
                parsed = json.loads(response.text)
                raise RuntimeError(f"critical_cmd_approve error: {parsed['error']}")
            else:
                raise RuntimeError("critical_cmd_approve failed")
        return
    except Exception as error:
        raise RuntimeError(f"critical_cmd_approve failed due to {repr(error)}") from error


def critical_cmd_reject(uuid: str, scope: str = OPENC3_SCOPE):
    try:
        endpoint = f"/openc3-api/criticalcmd/reject/{uuid}"
        response = openc3.script.API_SERVER.request("post", endpoint, json=True, scope=scope)
        if not response or response.status_code != 200:
            if response:
                parsed = json.loads(response.text)
                raise RuntimeError(f"critical_cmd_reject error: {parsed['error']}")
            else:
                raise RuntimeError("critical_cmd_reject failed")
        return
    except Exception as error:
        raise RuntimeError(f"critical_cmd_reject failed due to {repr(error)}") from error


def critical_cmd_can_approve(uuid: str, scope: str = OPENC3_SCOPE):
    try:
        endpoint = f"/openc3-api/criticalcmd/canapprove/{uuid}"
        response = openc3.script.API_SERVER.request("get", endpoint, scope=scope)
        if not response or response.status_code != 200:
            raise RuntimeError(f"Unexpected response to critical_cmd_can_approve: {repr(response)}")
        result = json.loads(response.text)
        if result["status"] == "ok":
            return True
        else:
            return False
    except Exception as error:
        raise RuntimeError(f"critical_cmd_approve failed due to {repr(error)}") from error



# Copyright 2023 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

from contextlib import contextmanager
from queue import SimpleQueue
from threading import Lock


class ConnectionPool:
    def __init__(self, ctor, pool_size):
        self.ctor = ctor
        self.count = 0
        self.pool_size = pool_size
        self.pool = SimpleQueue()
        self.lock = Lock()
        self.pipelines = {}

    @contextmanager
    def get(self):
        item = None
        with self.lock:
            if not self.pool.empty():
                item = self.pool.get(False)
            elif self.count < self.pool_size:
                item = self.ctor()
                self.count += 1
            else:
                item = self.pool.get()
        try:
            yield item
        finally:
            self.pool.put(item)



# Copyright 2023 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

from openc3.packets.packet import Packet
from openc3.utilities.logger import Logger


class PacketParser:
    @classmethod
    def parse_command(cls, parser, target_name, commands, warnings):
        parser = PacketParser(parser)
        parser.verify_parameters()
        return parser.create_command(target_name, commands, warnings)

    @classmethod
    def parse_telemetry(cls, parser, target_name, telemetry, latest_data, warnings):
        parser = PacketParser(parser)
        parser.verify_parameters()
        return parser.create_telemetry(target_name, telemetry, latest_data, warnings)

    @classmethod
    def check_item_data_types(cls, packet):
        try:
            for item in packet.sorted_items:
                item.check_default_and_range_data_types()

        except TypeError as error:
            # Add the target name and packet name to the error message so the user
            # can debug where the error occurred
            raise TypeError(f"{packet.target_name} {packet.packet_name} {error}") from error

    @classmethod
    def _check_for_duplicate(cls, type, list, packet):
        msg = None
        if list.get(packet.target_name):
            if list[packet.target_name].get(packet.packet_name):
                msg = f"{type} Packet {packet.target_name} {packet.packet_name} redefined."
                Logger.warn(msg)
        return msg

    @classmethod
    def _finish_create_command(cls, packet, commands, warnings):
        warning = PacketParser._check_for_duplicate("Command", commands, packet)
        if warning:
            warnings.append(warning)
        packet.define_reserved_items()
        if not commands.get(packet.target_name):
            commands[packet.target_name] = {}
        return packet

    @classmethod
    def _finish_create_telemetry(cls, packet, telemetry, latest_data, warnings):
        warning = PacketParser._check_for_duplicate("Telemetry", telemetry, packet)
        if warning:
            warnings.append(warning)
        packet.define_reserved_items()

        if not telemetry.get(packet.target_name):
            telemetry[packet.target_name] = {}
            latest_data[packet.target_name] = {}
        return packet

    def __init__(self, parser):
        self.parser = parser

    def verify_parameters(self):
        self.usage = f"{self.parser.keyword} <TARGET NAME> <PACKET NAME> <ENDIANNESS: BIG_ENDIAN/LITTLE_ENDIAN> <DESCRIPTION (Optional)>"
        self.parser.verify_num_parameters(3, 4, self.usage)
        self.parser.verify_parameter_naming(2)  # Packet name is the 2nd parameter

    def create_command(self, target_name, commands, warnings):
        packet = self._create_packet(target_name)
        return PacketParser._finish_create_command(packet, commands, warnings)

    def create_telemetry(self, target_name, telemetry, latest_data, warnings):
        packet = self._create_packet(target_name)
        return PacketParser._finish_create_telemetry(packet, telemetry, latest_data, warnings)

    def _create_packet(self, target_name):
        params = self.parser.parameters
        if target_name == "SYSTEM":
            target_name = params[0].upper()
        packet_name = params[1].upper()
        endianness = params[2].upper()
        # description is optional
        description = None
        if len(params) > 3:
            description = params[3]
        if endianness != "BIG_ENDIAN" and endianness != "LITTLE_ENDIAN":
            raise self.parser.error(
                f"Invalid endianness {params[2]}. Must be BIG_ENDIAN or LITTLE_ENDIAN.",
                self.usage,
            )
        return Packet(target_name, packet_name, endianness, description)



# Copyright 2024 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

from .accessor import Accessor
# HtmlAccessor uses lxml.html
import lxml.etree as ET


class XmlAccessor(Accessor):
    @classmethod
    def class_read_item(cls, item, buffer):
        if item.data_type == "DERIVED":
            return None
        doc = cls._buffer_to_doc(buffer)
        doc_value = doc.xpath(item.key)
        # The xpath method returns a boolean, float, string, or list of items
        # If it's a list then just grab the first value to convert
        if isinstance(doc_value, list):
            doc_value = doc_value[0]
        return cls.convert_to_type(doc_value, item)

    @classmethod
    def class_write_item(cls, item, value, buffer):
        if item.data_type == "DERIVED":
            return None
        doc = cls._buffer_to_doc(buffer)
        cls._update_doc(doc, item, value)
        buffer[0:] = cls._doc_to_buffer(doc)

    @classmethod
    def class_read_items(cls, items, buffer):
        doc = cls._buffer_to_doc(buffer)
        result = {}
        for item in items:
            if item.data_type == "DERIVED":
                result[item.name] = None
            else:
                doc_value = doc.xpath(item.key)
                # The xpath method returns a boolean, float, string, or list of items
                # If it's a list then just grab the first value to convert
                if isinstance(doc_value, list):
                    doc_value = doc_value[0]
                result[item.name] = cls.convert_to_type(doc_value, item)
        return result

    @classmethod
    def class_write_items(cls, items, values, buffer):
        doc = cls._buffer_to_doc(buffer)
        for index, item in enumerate(items):
            cls._update_doc(doc, item, values[index])
        buffer[0:] = cls._doc_to_buffer(doc)

    @classmethod
    def _buffer_to_doc(cls, buffer):
        return ET.fromstring(buffer.decode())

    @classmethod
    def _doc_to_buffer(cls, doc):
        return bytearray(ET.tostring(doc))

    @classmethod
    def _update_doc(cls, doc, item, value):
        # Split the attribute or text selector from the xpath so we get a
        # lxml.etree._Element that we can operate on. Otherwise we get a
        # lxml.etree._ElementUnicodeResult which is read-only.
        parts = item.key.split("/")
        path = "/".join(parts[0:-1])
        node = doc.xpath(path)[0]
        # Determine what the selector was trying to set
        if "@" in parts[-1]:
            node.attrib[parts[-1][1:]] = str(value)
        elif "text()" == parts[-1]:
            node.text = str(value)
        else:
            raise RuntimeError(f"Unknown selector: {item.key}")

    def enforce_encoding(self):
        return None

    def enforce_length(self):
        return False

    def enforce_short_buffer_allowed(self):
        return True

    def enforce_derived_write_conversion(self, item):
        return True



# Copyright 2025 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.
#
# A portion of this file was funded by Blue Origin Enterprises, L.P.
# See https://github.com/OpenC3/cosmos/pull/1953

import os
import tempfile
from openc3.config.config_parser import ConfigParser
from openc3.packets.packet import Packet
from openc3.packets.parsers.packet_parser import PacketParser
from openc3.packets.parsers.packet_item_parser import PacketItemParser
from openc3.packets.parsers.state_parser import StateParser
from openc3.packets.parsers.limits_parser import LimitsParser
from openc3.packets.parsers.limits_response_parser import LimitsResponseParser
from openc3.packets.parsers.format_string_parser import FormatStringParser
from openc3.packets.parsers.processor_parser import ProcessorParser
from openc3.conversions.generic_conversion import GenericConversion
from openc3.conversions.polynomial_conversion import PolynomialConversion
from openc3.conversions.segmented_polynomial_conversion import (
    SegmentedPolynomialConversion,
)
from openc3.utilities.extract import convert_to_value
from openc3.utilities.logger import Logger
from openc3.utilities.string import (
    filename_to_module,
    filename_to_class_name,
    class_name_to_filename,
)
from openc3.top_level import get_class_from_module


class PacketConfig:
    COMMAND = "Command"
    TELEMETRY = "Telemetry"

    def __init__(self):
        self.name = None
        self.telemetry = {}
        self.commands = {}
        self.limits_groups = {}
        self.limits_sets = ["DEFAULT"]
        # Hash of Hashes. First index by target name and then item name.
        # Returns an array of packets with that target and item.
        self.latest_data = {}
        self.warnings = []
        self.cmd_id_value_hash = {}
        self.tlm_id_value_hash = {}

        # Create unknown packets
        self.commands["UNKNOWN"] = {}
        self.commands["UNKNOWN"]["UNKNOWN"] = Packet("UNKNOWN", "UNKNOWN", "BIG_ENDIAN")
        self.telemetry["UNKNOWN"] = {}
        self.telemetry["UNKNOWN"]["UNKNOWN"] = Packet("UNKNOWN", "UNKNOWN", "BIG_ENDIAN")

        self.reset_processing_variables()

    #########################################################################
    # The following methods process a command or telemetry packet config file
    #########################################################################

    # Processes a OpenC3 configuration file and uses the keywords to build up
    # knowledge of the commands, telemetry, and limits groups.
    #
    # self.param filename [String] The name of the configuration file
    # self.param process_target_name [String] The target name. Pass None when parsing
    #   an xtce file to automatically determine the target name.
    def process_file(self, filename, process_target_name):
        # TODO: Handle .xtce files
        # if File.extname(filename).downcase == ".xtce"
        #   XtceParser.process(self.commands, self.telemetry, self.warnings, filename, process_target_name)
        #   return

        # Partial files are included into another file and thus aren't directly processed
        if os.path.basename(filename)[0] == "_":  # Partials start with underscore
            return

        self.converted_type = None
        self.converted_bit_size = None
        self.proc_text = ""
        self.building_generic_conversion = False

        process_target_name = process_target_name.upper()
        parser = ConfigParser("https://docs.openc3.com/docs")
        setattr(parser, "target_name", process_target_name)
        for keyword, params in parser.parse_file(filename):
            if self.building_generic_conversion:
                match keyword:
                    # Complete a generic conversion
                    case "GENERIC_READ_CONVERSION_END" | "GENERIC_WRITE_CONVERSION_END":
                        parser.verify_num_parameters(0, 0, keyword)
                        if "READ" in keyword:
                            self.current_item.read_conversion = GenericConversion(
                                self.proc_text,
                                self.converted_type,
                                self.converted_bit_size,
                            )
                        if "WRITE" in keyword:
                            self.current_item.write_conversion = GenericConversion(
                                self.proc_text,
                                self.converted_type,
                                self.converted_bit_size,
                            )
                        self.building_generic_conversion = False
                    # Add the current config.line to the conversion being built
                    case _:
                        self.proc_text += parser.line + "\n"

            else:  # not building generic conversion
                match keyword:
                    # Start a new packet
                    case "COMMAND":
                        self.finish_packet()
                        self.current_packet = PacketParser.parse_command(
                            parser, process_target_name, self.commands, self.warnings
                        )
                        self.current_cmd_or_tlm = PacketConfig.COMMAND

                    case "TELEMETRY":
                        self.finish_packet()
                        self.current_packet = PacketParser.parse_telemetry(
                            parser,
                            process_target_name,
                            self.telemetry,
                            self.latest_data,
                            self.warnings,
                        )
                        self.current_cmd_or_tlm = PacketConfig.TELEMETRY

                    # Select an existing packet for editing
                    case "SELECT_COMMAND" | "SELECT_TELEMETRY":
                        usage = f"{keyword} <TARGET NAME> <PACKET NAME>"
                        self.finish_packet()
                        parser.verify_num_parameters(2, 2, usage)
                        target_name = process_target_name
                        if target_name == "SYSTEM":
                            target_name = params[0].upper()
                        packet_name = params[1].upper()

                        self.current_packet = None
                        if "COMMAND" in keyword:
                            self.current_cmd_or_tlm = PacketConfig.COMMAND
                            if self.commands.get(target_name):
                                self.current_packet = self.commands[target_name][packet_name]
                        else:
                            self.current_cmd_or_tlm = PacketConfig.TELEMETRY
                            if self.telemetry.get(target_name):
                                self.current_packet = self.telemetry[target_name][packet_name]

                        if not self.current_packet:
                            raise parser.error("Packet not found", usage)

                    # Start the creation of a new limits group
                    case "LIMITS_GROUP":
                        usage = "LIMITS_GROUP <GROUP NAME>"
                        parser.verify_num_parameters(1, 1, usage)
                        self.current_limits_group = params[0].upper()
                        if self.current_limits_group not in self.limits_groups:
                            self.limits_groups[self.current_limits_group] = []

                    # Add a telemetry item to the limits group
                    case "LIMITS_GROUP_ITEM":
                        usage = "LIMITS_GROUP_ITEM <TARGET NAME> <PACKET NAME> <ITEM NAME>"
                        parser.verify_num_parameters(3, 3, usage)
                        if self.current_limits_group:
                            self.limits_groups[self.current_limits_group].append(
                                [
                                    params[0].upper(),
                                    params[1].upper(),
                                    params[2].upper(),
                                ]
                            )

                    #######################################################################
                    # All the following keywords must have a current packet defined
                    #######################################################################
                    case (
                        "SELECT_ITEM"
                        | "SELECT_PARAMETER"
                        | "DELETE_ITEM"
                        | "DELETE_PARAMETER"
                        | "ITEM"
                        | "PARAMETER"
                        | "ID_ITEM"
                        | "ID_PARAMETER"
                        | "ARRAY_ITEM"
                        | "ARRAY_PARAMETER"
                        | "APPEND_ITEM"
                        | "APPEND_PARAMETER"
                        | "APPEND_ID_ITEM"
                        | "APPEND_ID_PARAMETER"
                        | "APPEND_ARRAY_ITEM"
                        | "APPEND_ARRAY_PARAMETER"
                        | "ALLOW_SHORT"
                        | "HAZARDOUS"
                        | "PROCESSOR"
                        | "META"
                        | "DISABLE_MESSAGES"
                        | "HIDDEN"
                        | "DISABLED"
                        | "VIRTUAL"
                        | "ACCESSOR"
                        | "VALIDATOR"
                        | "TEMPLATE"
                        | "TEMPLATE_FILE"
                        | "RESPONSE"
                        | "ERROR_RESPONSE"
                        | "SCREEN"
                        | "RELATED_ITEM"
                        | "IGNORE_OVERLAP"
                    ):
                        if not self.current_packet:
                            raise parser.error(f"No current packet for {keyword}")
                        self.process_current_packet(parser, keyword, params)

                    #######################################################################
                    # All the following keywords must have a current item defined
                    #######################################################################
                    case (
                        "STATE"
                        | "READ_CONVERSION"
                        | "WRITE_CONVERSION"
                        | "POLY_READ_CONVERSION"
                        | "POLY_WRITE_CONVERSION"
                        | "SEG_POLY_READ_CONVERSION"
                        | "SEG_POLY_WRITE_CONVERSION"
                        | "GENERIC_READ_CONVERSION_START"
                        | "GENERIC_WRITE_CONVERSION_START"
                        | "REQUIRED"
                        | "LIMITS"
                        | "LIMITS_RESPONSE"
                        | "UNITS"
                        | "FORMAT_STRING"
                        | "DESCRIPTION"
                        | "MINIMUM_VALUE"
                        | "MAXIMUM_VALUE"
                        | "DEFAULT_VALUE"
                        | "OVERFLOW"
                        | "OVERLAP"
                        | "KEY"
                        | "VARIABLE_BIT_SIZE"
                    ):
                        if not self.current_item:
                            raise parser.error(f"No current item for {keyword}")
                        self.process_current_item(parser, keyword, params)

                    case _:
                        # blank config.lines will have a None keyword and should not raise an exception
                        if keyword:
                            raise parser.error(f"Unknown keyword '{keyword}'")

        # Complete the last defined packet
        self.finish_packet()

    # TODO: Used in xtce conversion
    # # Convert the PacketConfig back to OpenC3 configuration files for each target
    # def to_config(output_dir)
    #   FileUtils.mkdir_p(output_dir)

    #   self.telemetry.each do |target_name, packets|
    #     next if target_name == 'UNKNOWN'

    #     FileUtils.mkdir_p(File.join(output_dir, target_name, 'cmd_tlm'))
    #     filename = File.join(output_dir, target_name, 'cmd_tlm', target_name.downcase + '_tlm.txt')
    #     begin
    #       File.delete(filename)
    #     rescue
    #       # Doesn't exist

    #     packets.each do |packet_name, packet|
    #       File.open(filename, 'a') do |file|
    #         file.puts packet.to_config(:TELEMETRY)
    #         file.puts ""

    #   self.commands.each do |target_name, packets|
    #     next if target_name == 'UNKNOWN'

    #     FileUtils.mkdir_p(File.join(output_dir, target_name, 'cmd_tlm'))
    #     filename = File.join(output_dir, target_name, 'cmd_tlm', target_name.downcase + '_cmd.txt')
    #     begin
    #       File.delete(filename)
    #     rescue
    #       # Doesn't exist

    #     packets.each do |packet_name, packet|
    #       File.open(filename, 'a') do |file|
    #         file.puts packet.to_config(:COMMAND)
    #         file.puts ""

    #   # Put limits groups into SYSTEM target
    #   if len(self.limits_groups) > 0
    #     FileUtils.mkdir_p(File.join(output_dir, 'SYSTEM', 'cmd_tlm'))
    #     filename = File.join(output_dir, 'SYSTEM', 'cmd_tlm', 'limits_groups.txt')
    #     File.open(filename, 'w') do |file|
    #       self.limits_groups.each do |limits_group_name, limits_group_items|
    #         file.puts "LIMITS_GROUP {limits_group_name.quote_if_necessary}"
    #         limits_group_items.each do |target_name, packet_name, item_name|
    #           file.puts "  LIMITS_GROUP_ITEM {target_name.quote_if_necessary} {packet_name.quote_if_necessary} {item_name.quote_if_necessary}"

    #         file.puts ""

    #  # def to_config

    # def to_xtce(output_dir)
    #   XtceConverter.convert(self.commands, self.telemetry, output_dir)

    # Add current packet into hash if it exists
    def finish_packet(self):
        self.finish_item()
        if self.current_packet:
            warnings = self.current_packet.check_bit_offsets()
            if len(warnings) > 0:
                self.warnings += warnings
            if self.current_cmd_or_tlm == PacketConfig.COMMAND:
                PacketParser.check_item_data_types(self.current_packet)
                self.commands[self.current_packet.target_name][self.current_packet.packet_name] = self.current_packet
                if not self.current_packet.virtual:
                    id_values = self.cmd_id_value_hash.get(self.current_packet.target_name)
                    if not id_values:
                        id_values = {}
                    self.cmd_id_value_hash[self.current_packet.target_name] = id_values
                    self.update_id_value_hash(self.current_packet, id_values)
            else:
                self.telemetry[self.current_packet.target_name][self.current_packet.packet_name] = self.current_packet
                if not self.current_packet.virtual:
                    id_values = self.tlm_id_value_hash.get(self.current_packet.target_name)
                    if not id_values:
                        id_values = {}
                    self.tlm_id_value_hash[self.current_packet.target_name] = id_values
                    self.update_id_value_hash(self.current_packet, id_values)

            self.current_packet = None
            self.current_item = None

    def dynamic_add_packet(self, packet, cmd_or_tlm="TELEMETRY", affect_ids=False):
        if cmd_or_tlm == "COMMAND":
            self.commands[packet.target_name][packet.packet_name] = packet

            if affect_ids:
                id_values = self.cmd_id_value_hash.get(packet.target_name, None)
                if not id_values:
                    id_values = {}
                self.cmd_id_value_hash[packet.target_name] = id_values
                self.update_id_value_hash(packet, id_values)
        else:
            self.telemetry[packet.target_name][packet.packet_name] = packet

            # Update latest_data lookup for telemetry
            for item in packet.sorted_items:
                target_latest_data = self.latest_data[packet.target_name]
                if not target_latest_data.get(item.name, None):
                    target_latest_data[item.name] = []
                latest_data_packets = target_latest_data[item.name]
                if packet not in latest_data_packets:
                    latest_data_packets.append(packet)

            if affect_ids:
                id_values = self.tlm_id_value_hash.get(packet.target_name, None)
                if not id_values:
                    id_values = {}
                self.tlm_id_value_hash[packet.target_name] = id_values
                self.update_id_value_hash(packet, id_values)

    # This method provides way to quickly test packet configs
    #
    # from openc3.packets.packet_config import PacketConfig
    #
    # config = """
    #   ...
    # """
    #
    # pc = PacketConfig.from_config(config, "MYTARGET")
    # c = pc.commands['CMDADCS']['SET_POINTING_CMD']
    # c.restore_defaults()
    # c.write("MYITEM", 5)
    # print(c.buffer)
    @classmethod
    def from_config(cls, config, process_target_name):
        pc = cls()
        with tempfile.NamedTemporaryFile(mode="w+t", suffix=".txt") as tf:
            tf.write(config)
            tf.seek(0)
            pc.process_file(tf.name, process_target_name)
        return pc

    def update_id_value_hash(self, packet, hash):
        if packet.id_items and len(packet.id_items) > 0:
            key = []
            for item in packet.id_items:
                key.append(item.id_value)

            hash[repr(key)] = packet
        else:
            hash["CATCHALL"] = packet

    def reset_processing_variables(self):
        self.current_cmd_or_tlm = None
        self.current_packet = None
        self.current_item = None
        self.current_limits_group = None

    def process_current_packet(self, parser, keyword, params):
        match keyword:
            # Select or delete an item in the current packet
            case "SELECT_PARAMETER" | "SELECT_ITEM" | "DELETE_PARAMETER" | "DELETE_ITEM":
                if (self.current_cmd_or_tlm == PacketConfig.COMMAND) and (keyword.split("_")[1] == "ITEM"):
                    raise parser.error(f"{keyword} only applies to telemetry packets")
                if (self.current_cmd_or_tlm == PacketConfig.TELEMETRY) and (keyword.split("_")[1] == "PARAMETER"):
                    raise parser.error(f"{keyword} only applies to command packets")

                usage = f"{keyword} <{keyword.split('_')[1]} NAME>"
                self.finish_item()
                parser.verify_num_parameters(1, 1, usage)
                try:
                    if "SELECT" in keyword:
                        self.current_item = self.current_packet.get_item(params[0])
                    else:  # DELETE
                        self.current_packet.delete_item(params[0])
                except Exception:  # Rescue the default exception to provide a nicer error message
                    raise parser.error(
                        f"{params[0]} not found in {self.current_cmd_or_tlm.lower()} packet {self.current_packet.target_name} {self.current_packet.packet_name}",
                        usage,
                    )

            # Start a new telemetry item in the current packet
            case (
                "ITEM"
                | "PARAMETER"
                | "ID_ITEM"
                | "ID_PARAMETER"
                | "ARRAY_ITEM"
                | "ARRAY_PARAMETER"
                | "APPEND_ITEM"
                | "APPEND_PARAMETER"
                | "APPEND_ID_ITEM"
                | "APPEND_ID_PARAMETER"
                | "APPEND_ARRAY_ITEM"
                | "APPEND_ARRAY_PARAMETER"
            ):
                self.start_item(parser)

            # Allow this packet to be received with less data than the defined length
            # without generating a warning.
            case "ALLOW_SHORT":
                self.current_packet.short_buffer_allowed = True

            # Mark the current command as hazardous
            case "HAZARDOUS":
                usage = "HAZARDOUS <HAZARDOUS DESCRIPTION (Optional)>"
                parser.verify_num_parameters(0, 1, usage)
                self.current_packet.hazardous = True
                if len(params) == 1:
                    self.current_packet.hazardous_description = params[0]

            # Define a processor class that will be called once case a packet is received
            case "PROCESSOR":
                ProcessorParser.parse(parser, self.current_packet, self.current_cmd_or_tlm)

            case "DISABLE_MESSAGES":
                usage = keyword
                parser.verify_num_parameters(0, 0, usage)
                self.current_packet.messages_disabled = True

            # Store user defined metadata for the packet or a packet item
            case "META":
                usage = "META <META NAME> <META VALUES (optional)>"
                parser.verify_num_parameters(1, None, usage)
                if len(params) > 1:
                    meta_values = params[1:]
                else:
                    meta_values = []
                for index, value in enumerate(meta_values):
                    if isinstance(value, str):
                        meta_values[index] = value
                if self.current_item:
                    # Item META
                    self.current_item.meta[params[0].upper()] = meta_values
                else:
                    # Packet META
                    self.current_packet.meta[params[0].upper()] = meta_values

            case "HIDDEN":
                usage = keyword
                parser.verify_num_parameters(0, 0, usage)
                self.current_packet.hidden = True

            case "DISABLED":
                usage = keyword
                parser.verify_num_parameters(0, 0, usage)
                self.current_packet.hidden = True
                self.current_packet.disabled = True

            case "VIRTUAL":
                usage = keyword
                parser.verify_num_parameters(0, 0, usage)
                self.current_packet.hidden = True
                self.current_packet.disabled = True
                self.current_packet.virtual = True

            case "RESTRICTED":
                usage = keyword
                parser.verify_num_parameters(0, 0, usage)
                self.current_packet.restricted = True

            case "ACCESSOR":
                usage = f"{keyword} <File name> <Optional parameters> ..."
                parser.verify_num_parameters(1, None, usage)
                klass = None
                try:
                    klass = get_class_from_module(
                        filename_to_module(params[0]),
                        filename_to_class_name(params[0]),
                    )
                except ModuleNotFoundError:
                    try:
                        # Fall back to the deprecated behavior of passing the ClassName (only works with built-in accessors)
                        filename = class_name_to_filename(params[0])
                        klass = get_class_from_module(f"openc3.accessors.{filename}", params[0])
                    except ModuleNotFoundError:
                        raise parser.error(f"ModuleNotFoundError parsing {params[0]}. Usage: {usage}")
                if len(params) > 1:
                    self.current_packet.accessor = klass(self.current_packet, *params[1:])
                else:
                    self.current_packet.accessor = klass(self.current_packet)

            case "VALIDATOR":
                usage = f"{keyword} <Class name> <Optional parameters> ..."
                parser.verify_num_parameters(1, None, usage)
                try:
                    klass = get_class_from_module(
                        filename_to_module(params[0]),
                        filename_to_class_name(params[0]),
                    )
                    if len(params) > 1:
                        self.current_packet.validator = klass(self.current_packet, *params[1:])
                    else:
                        self.current_packet.validator = klass(self.current_packet)
                except ModuleNotFoundError as error:
                    raise parser.error(error)

            case "TEMPLATE":
                usage = f"{keyword} <Template string>"
                parser.verify_num_parameters(1, 1, usage)
                self.current_packet.template = bytearray(params[0], "ascii")

            case "TEMPLATE_FILE":
                usage = f"{keyword} <Template file path>"
                parser.verify_num_parameters(1, 1, usage)

                try:
                    self.current_packet.template = parser.read_file(params[0])
                except OSError as error:
                    raise parser.error(error)

            case "RESPONSE":
                usage = f"{keyword} <Target Name> <Packet Name>"
                parser.verify_num_parameters(2, 2, usage)
                if self.current_cmd_or_tlm == PacketConfig.TELEMETRY:
                    raise parser.error(f"{keyword} only applies to command packets")
                self.current_packet.response = [params[0].upper(), params[1].upper()]

            case "ERROR_RESPONSE":
                usage = f"{keyword} <Target Name> <Packet Name>"
                parser.verify_num_parameters(2, 2, usage)
                if self.current_cmd_or_tlm == PacketConfig.TELEMETRY:
                    raise parser.error(f"{keyword} only applies to command packets")
                self.current_packet.error_response = [
                    params[0].upper(),
                    params[1].upper(),
                ]

            case "SCREEN":
                usage = f"{keyword} <Target Name> <Screen Name>"
                parser.verify_num_parameters(2, 2, usage)
                if self.current_cmd_or_tlm == PacketConfig.TELEMETRY:
                    raise parser.error(f"{keyword} only applies to command packets")
                self.current_packet.screen = [params[0].upper(), params[1].upper()]

            case "RELATED_ITEM":
                usage = f"{keyword} <Target Name> <Packet Name> <Item Name>"
                parser.verify_num_parameters(3, 3, usage)
                if self.current_cmd_or_tlm == PacketConfig.TELEMETRY:
                    raise parser.error(f"{keyword} only applies to command packets")
                if not self.current_packet.related_items:
                    self.current_packet.related_items = []
                self.current_packet.related_items.append([params[0].upper(), params[1].upper(), params[2].upper()])

            case "IGNORE_OVERLAP":
                usage = f"{keyword}"
                parser.verify_num_parameters(0, 0, usage)
                self.current_packet.ignore_overlap = True

    def process_current_item(self, parser, keyword, params):
        match keyword:
            # Add a state to the current telemety item
            case "STATE":
                StateParser.parse(
                    parser,
                    self.current_packet,
                    self.current_cmd_or_tlm,
                    self.current_item,
                    self.warnings,
                )

            # Apply a conversion to the current item after it is read to or
            # written from the packet
            case "READ_CONVERSION" | "WRITE_CONVERSION":
                usage = f"{keyword} <conversion class filename> <custom parameters> ..."
                parser.verify_num_parameters(1, None, usage)
                try:
                    klass = get_class_from_module(
                        filename_to_module(params[0]),
                        filename_to_class_name(params[0]),
                    )
                    conversion = klass(*params[1:])
                    if "READ" in keyword:
                        self.current_item.read_conversion = conversion
                    else:
                        self.current_item.write_conversion = conversion
                    # if klass != ProcessorConversion and (conversion.converted_type.None? or conversion.converted_bit_size.None?):
                    #     msg = f"Read Conversion {params[0]} on item {self.current_item.name} does not specify converted type or bit size"
                    #     self.warnings.append(msg)
                    #     Logger.warn(self.warnings[-1])
                except Exception as error:
                    raise parser.error(error)

            # Apply a polynomial conversion to the current item
            case "POLY_READ_CONVERSION" | "POLY_WRITE_CONVERSION":
                usage = f"{keyword} <C0> <C1> <C2> ..."
                parser.verify_num_parameters(1, None, usage)
                if "READ" in keyword:
                    self.current_item.read_conversion = PolynomialConversion(*params)
                else:
                    self.current_item.write_conversion = PolynomialConversion(*params)

            # Apply a segmented polynomial conversion to the current item
            # after it is read from the telemetry packet
            case "SEG_POLY_READ_CONVERSION":
                usage = "SEG_POLY_READ_CONVERSION <Lower Bound> <C0> <C1> <C2> ..."
                parser.verify_num_parameters(2, None, usage)
                if (
                    self.current_item.read_conversion is None
                    or self.current_item.read_conversion.__class__ != SegmentedPolynomialConversion
                ):
                    self.current_item.read_conversion = SegmentedPolynomialConversion()
                self.current_item.read_conversion.add_segment(float(params[0]), *params[1:])

            # Apply a segmented polynomial conversion to the current item
            # before it is written to the telemetry packet
            case "SEG_POLY_WRITE_CONVERSION":
                usage = "SEG_POLY_WRITE_CONVERSION <Lower Bound> <C0> <C1> <C2> ..."
                parser.verify_num_parameters(2, None, usage)
                if (
                    self.current_item.write_conversion is None
                    or self.current_item.write_conversion.__class__ != SegmentedPolynomialConversion
                ):
                    self.current_item.write_conversion = SegmentedPolynomialConversion()
                self.current_item.write_conversion.add_segment(float(params[0]), *params[1:])

            # Start the definition of a generic conversion.
            # All config.lines following this config.line are considered part
            # of the conversion until an end of conversion marker is found
            case "GENERIC_READ_CONVERSION_START" | "GENERIC_WRITE_CONVERSION_START":
                usage = f"{keyword} <Converted Type (optional)> <Converted Bit Size (optional)>"
                parser.verify_num_parameters(0, 2, usage)
                self.proc_text = ""
                self.building_generic_conversion = True
                self.converted_type = None
                self.converted_bit_size = None
                if len(params) == 2:
                    self.converted_type = params[0].upper()
                    if self.converted_type not in [
                        "INT",
                        "UINT",
                        "FLOAT",
                        "STRING",
                        "BLOCK",
                    ]:
                        raise parser.error(f"Invalid converted_type: {self.converted_type}.")
                    self.converted_bit_size = int(params[1])
                if self.converted_type is None or self.converted_bit_size is None:
                    msg = f"Generic Conversion on item {self.current_item.name} does not specify converted type or bit size"
                    self.warnings.append(msg)
                    Logger.warn(self.warnings[-1])

            # Define a set of limits for the current telemetry item
            case "LIMITS":
                self.limits_sets.append(
                    LimitsParser.parse(
                        parser,
                        self.current_packet,
                        self.current_cmd_or_tlm,
                        self.current_item,
                        self.warnings,
                    )
                )
                self.limits_sets = list(set(self.limits_sets))

            # Define a response class that will be called case the limits state of the:
            # current item changes.
            case "LIMITS_RESPONSE":
                LimitsResponseParser.parse(parser, self.current_item, self.current_cmd_or_tlm)

            # Define a printf style formatting string for the current telemetry item
            case "FORMAT_STRING":
                FormatStringParser.parse(parser, self.current_item)

            # Define the units of the current telemetry item
            case "UNITS":
                usage = "UNITS <FULL UNITS NAME> <ABBREVIATED UNITS NAME>"
                parser.verify_num_parameters(2, 2, usage)
                self.current_item.units_full = params[0]
                self.current_item.units = params[1]

            # Update the description for the current telemetry item
            case "DESCRIPTION":
                usage = "DESCRIPTION <DESCRIPTION>"
                parser.verify_num_parameters(1, 1, usage)
                self.current_item.description = params[0]

            # Mark the current command parameter as required.
            # This means it must be given a value and not just use its default.
            case "REQUIRED":
                usage = "REQUIRED"
                parser.verify_num_parameters(0, 0, usage)
                if self.current_cmd_or_tlm == PacketConfig.COMMAND:
                    self.current_item.required = True
                else:
                    raise parser.error(f"{keyword} only applies to command parameters")

            # Update the minimum value for the current command parameter
            case "MINIMUM_VALUE":
                if self.current_cmd_or_tlm == PacketConfig.TELEMETRY:
                    raise parser.error(f"{keyword} only applies to command parameters")

                usage = "MINIMUM_VALUE <MINIMUM VALUE>"
                parser.verify_num_parameters(1, 1, usage)
                min = ConfigParser.handle_defined_constants(
                    convert_to_value(params[0]),
                    self.current_item.data_type,
                    self.current_item.bit_size,
                )
                self.current_item.minimum = min

            # Update the maximum value for the current command parameter
            case "MAXIMUM_VALUE":
                if self.current_cmd_or_tlm == PacketConfig.TELEMETRY:
                    raise parser.error(f"{keyword} only applies to command parameters")

                usage = "MAXIMUM_VALUE <MAXIMUM VALUE>"
                parser.verify_num_parameters(1, 1, usage)
                max = ConfigParser.handle_defined_constants(
                    convert_to_value(params[0]),
                    self.current_item.data_type,
                    self.current_item.bit_size,
                )
                self.current_item.maximum = max

            # Update the default value for the current command parameter
            case "DEFAULT_VALUE":
                if self.current_cmd_or_tlm == PacketConfig.TELEMETRY:
                    raise parser.error(f"{keyword} only applies to command parameters")

                usage = "DEFAULT_VALUE <DEFAULT VALUE>"
                parser.verify_num_parameters(1, 1, usage)
                if (self.current_item.data_type == "STRING") or (self.current_item.data_type == "BLOCK"):
                    self.current_item.default = params[0]
                else:
                    self.current_item.default = ConfigParser.handle_defined_constants(
                        convert_to_value(params[0]),
                        self.current_item.data_type,
                        self.current_item.bit_size,
                    )

            # Update the overflow type for the current command parameter
            case "OVERFLOW":
                usage = "OVERFLOW <OVERFLOW VALUE - ERROR, ERROR_ALLOW_HEX, TRUNCATE, or SATURATE>"
                parser.verify_num_parameters(1, 1, usage)
                self.current_item.overflow = params[0].upper()

            case "OVERLAP":
                parser.verify_num_parameters(0, 0, "OVERLAP")
                self.current_item.overlap = True

            case "KEY":
                parser.verify_num_parameters(1, 1, "KEY <key or path into data>")
                self.current_item.key = params[0]

            case "VARIABLE_BIT_SIZE":
                parser.verify_num_parameters(
                    1,
                    3,
                    "VARIABLE_BIT_SIZE <length_item_name> <length_bits_per_count = 8> <length_value_bit_offset = 0>",
                )

                variable_bit_size = {"length_bits_per_count": 8, "length_value_bit_offset": 0}
                variable_bit_size["length_item_name"] = params[0].upper()
                if len(params) > 1:
                    variable_bit_size["length_bits_per_count"] = int(params[1])
                if len(params) > 2:
                    variable_bit_size["length_value_bit_offset"] = int(params[2])
                self.current_item.variable_bit_size = variable_bit_size

    def start_item(self, parser):
        self.finish_item()
        self.current_item = PacketItemParser.parse(parser, self.current_packet, self.current_cmd_or_tlm, self.warnings)

    # Finish updating packet item
    def finish_item(self):
        if self.current_item:
            self.current_packet.set_item(self.current_item)
            if self.current_cmd_or_tlm == PacketConfig.TELEMETRY:
                target_latest_data = self.latest_data[self.current_packet.target_name]
                if not target_latest_data.get(self.current_item.name):
                    target_latest_data[self.current_item.name] = []
                latest_data_packets = target_latest_data[self.current_item.name]
                if self.current_packet not in latest_data_packets:
                    latest_data_packets.append(self.current_packet)
            self.current_item = None



# Copyright 2023 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

import json
from openc3.topics.topic import Topic
from openc3.environment import OPENC3_SCOPE


class InterfaceTopic(Topic):
    while_receive_commands = False

    # Generate a list of topics for this interface. This includes the interface itself
    # and all the targets which are assigned to this interface.
    @classmethod
    def topics(cls, interface, scope=OPENC3_SCOPE):
        topics = []
        topics.append(f"{{{scope}__CMD}}INTERFACE__{interface.name}")
        for target_name in interface.cmd_target_names:
            topics.append(f"{{{scope}__CMD}}TARGET__{target_name}")
        topics.append("OPENC3__SYSTEM__EVENTS")  # Add System Events
        return topics

    @classmethod
    def receive_commands(cls, method, interface, scope=OPENC3_SCOPE):
        InterfaceTopic.while_receive_commands = True
        while InterfaceTopic.while_receive_commands:
            for topic, msg_id, msg_hash, redis in Topic.read_topics(InterfaceTopic.topics(interface, scope)):
                result = method(topic, msg_id, msg_hash, redis)
                ack_topic = topic.split("__")
                ack_topic[1] = "ACK" + ack_topic[1]
                ack_topic = "__".join(ack_topic)
                Topic.write_topic(ack_topic, {"result": result, "id": msg_id}, "*", 100)

    @classmethod
    def write_raw(cls, interface_name, data, scope):
        Topic.write_topic(f"{{{scope}__CMD}}INTERFACE__{interface_name}", {"raw": data}, "*", 100)
        # TODO: This should wait for the ack

    @classmethod
    def connect_interface(cls, interface_name, *interface_params, scope=OPENC3_SCOPE):
        if interface_params and len(interface_params) != 0:
            Topic.write_topic(
                f"{{{scope}__CMD}}INTERFACE__{interface_name}",
                {"connect": "true", "params": json.dumps(interface_params)},
                "*",
                100,
            )
        else:
            Topic.write_topic(
                f"{{{scope}__CMD}}INTERFACE__{interface_name}",
                {"connect": "true"},
                "*",
                100,
            )

    @classmethod
    def disconnect_interface(cls, interface_name, scope=OPENC3_SCOPE):
        Topic.write_topic(
            f"{{{scope}__CMD}}INTERFACE__{interface_name}",
            {"disconnect": "true"},
            "*",
            100,
        )

    @classmethod
    def start_raw_logging(cls, interface_name, scope=OPENC3_SCOPE):
        Topic.write_topic(
            f"{{{scope}__CMD}}INTERFACE__{interface_name}",
            {"log_stream": "true"},
            "*",
            100,
        )

    @classmethod
    def stop_raw_logging(cls, interface_name, scope=OPENC3_SCOPE):
        Topic.write_topic(
            f"{{{scope}__CMD}}INTERFACE__{interface_name}",
            {"log_stream": "false"},
            "*",
            100,
        )

    @classmethod
    def shutdown(cls, interface, scope=OPENC3_SCOPE):
        InterfaceTopic.while_receive_commands = False
        Topic.write_topic(
            f"{{{scope}__CMD}}INTERFACE__{interface.name}",
            {"shutdown": "true"},
            "*",
            100,
        )

    @classmethod
    def interface_cmd(cls, interface_name, cmd_name, *cmd_params, scope=OPENC3_SCOPE):
        data = {}
        data["cmd_name"] = cmd_name
        data["cmd_params"] = cmd_params
        Topic.write_topic(
            f"{{{scope}__CMD}}INTERFACE__{interface_name}",
            {"interface_cmd": json.dumps(data)},
            "*",
            100,
        )

    @classmethod
    def protocol_cmd(
        cls,
        interface_name,
        cmd_name,
        *cmd_params,
        read_write="READ_WRITE",
        index=-1,
        scope=OPENC3_SCOPE,
    ):
        data = {}
        data["cmd_name"] = cmd_name
        data["cmd_params"] = cmd_params
        data["read_write"] = str(read_write).upper()
        data["index"] = index
        Topic.write_topic(
            f"{{{scope}__CMD}}INTERFACE__{interface_name}",
            {"protocol_cmd": json.dumps(data)},
            "*",
            100,
        )

    @classmethod
    def inject_tlm(
        cls,
        interface_name,
        target_name,
        packet_name,
        item_hash=None,
        type="CONVERTED",
        scope=OPENC3_SCOPE,
    ):
        data = {}
        data["target_name"] = target_name.upper()
        data["packet_name"] = packet_name.upper()
        data["item_hash"] = item_hash
        data["type"] = type
        Topic.write_topic(
            f"{{{scope}__CMD}}INTERFACE__{interface_name}",
            {"inject_tlm": json.dumps(data)},
            "*",
            100,
        )



# Copyright 2023 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

import json
from openc3.topics.topic import Topic
from openc3.models.cvt_model import CvtModel
from openc3.utilities.json import JsonEncoder
from openc3.utilities.time import to_nsec_from_epoch


class TelemetryDecomTopic(Topic):
    @classmethod
    def write_packet(cls, packet, id=None, scope=None):
        # OpenC3.in_span("write_packet") do
        # Need to build a JSON hash of the decommutated data
        # Support "downward typing"
        # everything base name is RAW (including DERIVED)
        # Request for WITH_UNITS, etc will look down until it finds something
        # If nothing - item does not exist - nil
        # __ as separators ITEM1, ITEM1__C, ITEM1__F, ITEM1__U

        json_hash = CvtModel.build_json_from_packet(packet)
        # Write to stream
        msg_hash = {
            "time": to_nsec_from_epoch(packet.packet_time),
            "stored": str(packet.stored),
            "target_name": packet.target_name,
            "packet_name": packet.packet_name,
            "received_count": packet.received_count,
            "json_data": json.dumps(json_hash, cls=JsonEncoder),
        }
        Topic.write_topic(
            f"{scope}__DECOM__{{{packet.target_name}}}__{packet.packet_name}",
            msg_hash,
            id,
        )

        if not packet.stored:
            # Also update the current value table with the latest decommutated data
            CvtModel.set(
                json_hash,
                packet.target_name,
                packet.packet_name,
                scope=scope,
            )



# Copyright 2024 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

import json
from openc3.topics.topic import Topic
from openc3.utilities.store_queued import EphemeralStoreQueued
from openc3.environment import OPENC3_SCOPE
from openc3.utilities.json import JsonEncoder
from openc3.utilities.time import to_nsec_from_epoch


class CommandDecomTopic(Topic):
    @classmethod
    def topics(cls, scope):
        Topic.topics("DECOMCMD", scope)

    @classmethod
    def write_packet(cls, packet, scope):
        topic = f"{scope}__DECOMCMD__{{{packet.target_name}}}__{packet.packet_name}"
        msg_hash = {
            "time": to_nsec_from_epoch(packet.packet_time),
            "target_name": packet.target_name,
            "packet_name": packet.packet_name,
            "stored": str(packet.stored),
            "received_count": packet.received_count,
        }
        json_hash = {}
        for item in packet.sorted_items:
            json_hash[item.name] = packet.read_item(item, "RAW")
            if item.write_conversion or item.states:
                json_hash[item.name + "__C"] = packet.read_item(item, "CONVERTED")
            if item.format_string:
                json_hash[item.name + "__F"] = packet.read_item(item, "FORMATTED")
            if item.units:
                json_hash[item.name + "__U"] = packet.read_item(item, "WITH_UNITS")
        json_hash["extra"] = json.dumps(packet.extra, cls=JsonEncoder)
        msg_hash["json_data"] = json.dumps(json_hash, cls=JsonEncoder)
        EphemeralStoreQueued.write_topic(topic, msg_hash)

    @classmethod
    def get_cmd_item(cls, target_name, packet_name, param_name, type="WITH_UNITS", scope=OPENC3_SCOPE):
        msg_id, msg_hash = Topic.get_newest_message(f"{scope}__DECOMCMD__{{{target_name}}}__{packet_name}")
        if msg_id:
            if param_name == "RECEIVED_COUNT":
                return int(msg_hash[b"received_count"])
            else:
                cmd_item = json.loads(msg_hash[b"json_data"])
                # Start from the most complex down to the basic raw value
                value = cmd_item.get(f"{param_name}__U")
                if value is not None and type == "WITH_UNITS":
                    return value

                value = cmd_item.get(f"{param_name}__F")
                if value is not None and (type == "WITH_UNITS" or type == "FORMATTED"):
                    return value

                value = cmd_item.get(f"{param_name}__C")
                if value is not None and (type == "WITH_UNITS" or type == "FORMATTED" or type == "CONVERTED"):
                    return value

                return cmd_item[param_name]



# Copyright 2023 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

from openc3.top_level import get_class_from_module
from openc3.utilities.string import filename_to_module, filename_to_class_name


class LimitsResponseParser:
    # self.param parser [ConfigParser] Configuration parser
    # self.param item [Packet] The current item
    # self.param cmd_or_tlm [String] Whether this is a command or telemetry packet
    @classmethod
    def parse(cls, parser, item, cmd_or_tlm):
        parser = LimitsResponseParser(parser)
        parser.verify_parameters(cmd_or_tlm)
        parser.create_limits_response(item)

    # self.param parser [ConfigParser] Configuration parser
    def __init__(self, parser):
        self.parser = parser

    # self.param cmd_or_tlm [String] Whether this is a command or telemetry packet
    def verify_parameters(self, cmd_or_tlm):
        if cmd_or_tlm == "Command":
            raise self.parser.error("LIMITS_RESPONSE only applies to telemetry items")

        self.usage = "LIMITS_RESPONSE <RESPONSE CLASS FILENAME> <RESPONSE SPECIFIC OPTIONS>"
        self.parser.verify_num_parameters(1, None, self.usage)

    # self.param item [PacketItem] The item the limits response should be added to
    def create_limits_response(self, item):
        try:
            klass = get_class_from_module(
                filename_to_module(self.parser.parameters[0]),
                filename_to_class_name(self.parser.parameters[0]),
            )
            if len(self.parser.parameters) > 1:
                item.limits.response = klass(*self.parser.parameters[1 : len(self.parser.parameters)])
            else:
                item.limits.response = klass()
        except ModuleNotFoundError as error:
            raise self.parser.error(error, self.usage)



# Copyright 2022 Ball Aerospace & Technologies Corp.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt

# Modified by OpenC3, Inc.
# All changes Copyright 2023, OpenC3, Inc.
# All Rights Reserved
#
# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

import struct


# Abstract base class which {Crc16}, {Crc32} and {Crc64} use. Do NOT use this
# class directly but instead use one of the subclasses.
class Crc:
    BIT_REVERSE_TABLE = [
        0x00,
        0x80,
        0x40,
        0xC0,
        0x20,
        0xA0,
        0x60,
        0xE0,
        0x10,
        0x90,
        0x50,
        0xD0,
        0x30,
        0xB0,
        0x70,
        0xF0,
        0x08,
        0x88,
        0x48,
        0xC8,
        0x28,
        0xA8,
        0x68,
        0xE8,
        0x18,
        0x98,
        0x58,
        0xD8,
        0x38,
        0xB8,
        0x78,
        0xF8,
        0x04,
        0x84,
        0x44,
        0xC4,
        0x24,
        0xA4,
        0x64,
        0xE4,
        0x14,
        0x94,
        0x54,
        0xD4,
        0x34,
        0xB4,
        0x74,
        0xF4,
        0x0C,
        0x8C,
        0x4C,
        0xCC,
        0x2C,
        0xAC,
        0x6C,
        0xEC,
        0x1C,
        0x9C,
        0x5C,
        0xDC,
        0x3C,
        0xBC,
        0x7C,
        0xFC,
        0x02,
        0x82,
        0x42,
        0xC2,
        0x22,
        0xA2,
        0x62,
        0xE2,
        0x12,
        0x92,
        0x52,
        0xD2,
        0x32,
        0xB2,
        0x72,
        0xF2,
        0x0A,
        0x8A,
        0x4A,
        0xCA,
        0x2A,
        0xAA,
        0x6A,
        0xEA,
        0x1A,
        0x9A,
        0x5A,
        0xDA,
        0x3A,
        0xBA,
        0x7A,
        0xFA,
        0x06,
        0x86,
        0x46,
        0xC6,
        0x26,
        0xA6,
        0x66,
        0xE6,
        0x16,
        0x96,
        0x56,
        0xD6,
        0x36,
        0xB6,
        0x76,
        0xF6,
        0x0E,
        0x8E,
        0x4E,
        0xCE,
        0x2E,
        0xAE,
        0x6E,
        0xEE,
        0x1E,
        0x9E,
        0x5E,
        0xDE,
        0x3E,
        0xBE,
        0x7E,
        0xFE,
        0x01,
        0x81,
        0x41,
        0xC1,
        0x21,
        0xA1,
        0x61,
        0xE1,
        0x11,
        0x91,
        0x51,
        0xD1,
        0x31,
        0xB1,
        0x71,
        0xF1,
        0x09,
        0x89,
        0x49,
        0xC9,
        0x29,
        0xA9,
        0x69,
        0xE9,
        0x19,
        0x99,
        0x59,
        0xD9,
        0x39,
        0xB9,
        0x79,
        0xF9,
        0x05,
        0x85,
        0x45,
        0xC5,
        0x25,
        0xA5,
        0x65,
        0xE5,
        0x15,
        0x95,
        0x55,
        0xD5,
        0x35,
        0xB5,
        0x75,
        0xF5,
        0x0D,
        0x8D,
        0x4D,
        0xCD,
        0x2D,
        0xAD,
        0x6D,
        0xED,
        0x1D,
        0x9D,
        0x5D,
        0xDD,
        0x3D,
        0xBD,
        0x7D,
        0xFD,
        0x03,
        0x83,
        0x43,
        0xC3,
        0x23,
        0xA3,
        0x63,
        0xE3,
        0x13,
        0x93,
        0x53,
        0xD3,
        0x33,
        0xB3,
        0x73,
        0xF3,
        0x0B,
        0x8B,
        0x4B,
        0xCB,
        0x2B,
        0xAB,
        0x6B,
        0xEB,
        0x1B,
        0x9B,
        0x5B,
        0xDB,
        0x3B,
        0xBB,
        0x7B,
        0xFB,
        0x07,
        0x87,
        0x47,
        0xC7,
        0x27,
        0xA7,
        0x67,
        0xE7,
        0x17,
        0x97,
        0x57,
        0xD7,
        0x37,
        0xB7,
        0x77,
        0xF7,
        0x0F,
        0x8F,
        0x4F,
        0xCF,
        0x2F,
        0xAF,
        0x6F,
        0xEF,
        0x1F,
        0x9F,
        0x5F,
        0xDF,
        0x3F,
        0xBF,
        0x7F,
        0xFF,
    ]

    # Creates a CRC algorithm instance.
    #
    # self.param poly [Integer] Polynomial to use when calculating the CRC
    # self.param seed [Integer] Seed value to start the calculation
    # self.param xor [Boolean] Whether to XOR the CRC result with 0xFFFF
    # self.param reflect [Boolean] Whether to bit reverse each byte of data before
    #   calculating the CRC
    def __init__(self, poly, seed, xor, reflect):
        self.poly = poly
        self.seed = seed
        self.xor = xor
        self.reflect = reflect
        self.table = []

        # Determine which class we're using= Crc8, Crc16, Crc32, Crc64
        match self.__class__.__name__:
            case "Crc8":
                self.bit_size = 8
                pack = ">B"
                filter_mask = 0xFF
            case "Crc16":
                self.bit_size = 16
                pack = ">H"
                filter_mask = 0xFFFF
            case "Crc32":
                self.bit_size = 32
                pack = ">I"
                filter_mask = 0xFFFFFFFF
            case "Crc64":
                self.bit_size = 64
                pack = ">Q"
                filter_mask = 0xFFFFFFFFFFFFFFFF
        for index in range(0, 256):
            self.table.append(
                int.from_bytes(
                    struct.pack(
                        pack,
                        self.compute_table_entry(index, self.bit_size) & filter_mask,
                    ),
                    byteorder="big",
                )
            )

    # self.!method calc(data, seed = None)
    #   Calculates the CRC across the data buffer using the optional seed.
    #   Implemented in C for speed.
    #
    #   self.param data [String] String buffer of binary data to calculate a CRC on
    #   self.param seed [Integer|None] Seed value to start the calculation. Pass None
    #     to use the default seed set in the constructor.
    #   self.return [Integer] The CRC value

    # Bit reverse the 8 bit value
    # self.param value [Integer]
    # self.return [Integer] Bit reversed value
    def bit_reverse_8(self, value):
        return Crc.BIT_REVERSE_TABLE[value & 0xFF]

    def bit_reverse_16(self, value):
        return (Crc.BIT_REVERSE_TABLE[value & 0xFF] << 8) | (Crc.BIT_REVERSE_TABLE[(value >> 8) & 0xFF])

    def bit_reverse_32(self, value):
        return (
            (Crc.BIT_REVERSE_TABLE[value & 0xFF] << 24)
            | (Crc.BIT_REVERSE_TABLE[(value >> 8) & 0xFF] << 16)
            | (Crc.BIT_REVERSE_TABLE[(value >> 16) & 0xFF] << 8)
            | (Crc.BIT_REVERSE_TABLE[(value >> 24) & 0xFF])
        )

    def bit_reverse_64(self, value):
        return (
            (Crc.BIT_REVERSE_TABLE[value & 0x00000000000000FF] << 56)
            | (Crc.BIT_REVERSE_TABLE[(value >> 8) & 0x00000000000000FF] << 48)
            | (Crc.BIT_REVERSE_TABLE[(value >> 16) & 0x00000000000000FF] << 40)
            | (Crc.BIT_REVERSE_TABLE[(value >> 24) & 0x00000000000000FF] << 32)
            | (Crc.BIT_REVERSE_TABLE[(value >> 32) & 0x00000000000000FF] << 24)
            | (Crc.BIT_REVERSE_TABLE[(value >> 40) & 0x00000000000000FF] << 16)
            | (Crc.BIT_REVERSE_TABLE[(value >> 48) & 0x00000000000000FF] << 8)
            | (Crc.BIT_REVERSE_TABLE[(value >> 56) & 0x00000000000000FF])
        )

    def calc(self, data, seed=None):
        if seed is None:
            seed = self.seed
        crc = seed

        match self.bit_size:
            case 8:
                right_shift = 0
                filter_mask = 0xFF
                final_bit_reverse = self.bit_reverse_8
            case 16:
                right_shift = 8
                filter_mask = 0xFFFF
                final_bit_reverse = self.bit_reverse_16
            case 32:
                right_shift = 24
                filter_mask = 0xFFFFFFFF
                final_bit_reverse = self.bit_reverse_32
            case 64:
                right_shift = 56
                filter_mask = 0xFFFFFFFFFFFFFFFF
                final_bit_reverse = self.bit_reverse_64

        if self.reflect:
            for byte in data:
                if isinstance(byte, str):
                    byte = ord(byte)
                crc = (crc << 8 & filter_mask) ^ self.table[(crc >> right_shift) ^ self.bit_reverse_8(byte)]

            final_bit_reverse(crc ^ filter_mask)
            if self.xor:
                return final_bit_reverse(crc ^ filter_mask)
            else:
                return final_bit_reverse(crc)
        else:
            for byte in data:
                if isinstance(byte, str):
                    byte = ord(byte)
                crc = ((crc << 8) & filter_mask) ^ self.table[(crc >> right_shift) ^ byte]

            if self.xor:
                return crc ^ filter_mask
            else:
                return crc

    # Compute a single entry in the crc lookup table
    def compute_table_entry(self, index, digits):
        # Start by shifting the index
        crc = index << (digits - 8)

        # The mask is 0x8000 for Crc16, 0x80000000 for Crc32, etc
        mask = 1 << (digits - 1)

        for _ in range(0, 8):
            if (crc & mask) != 0:
                crc = crc << 1 ^ self.poly
            else:
                crc = crc << 1

        # XOR the mask and or back in the top bit to get all ones
        mask = ~mask | mask
        return crc & mask


# Calculates 8-bit CRCs over a buffer of data.
class Crc8(Crc):
    # CRC-8-DVB-S2 default polynomial
    DEFAULT_POLY = 0xD5
    # Seed for 8-bit CRC
    DEFAULT_SEED = 0x00

    # Creates a 8 bit CRC algorithm instance. By default it is initialized to
    # use the CRC-8-DVB-S2 algorithm.
    #
    # self.param poly [Integer] Polynomial to use when calculating the CRC
    # self.param seed [Integer] Seed value to start the calculation
    # self.param xor [Boolean] Whether to XOR the CRC result with 0xFF
    # self.param reflect [Boolean] Whether to bit reverse each byte of data before
    #   calculating the CRC
    def __init__(self, poly=DEFAULT_POLY, seed=DEFAULT_SEED, xor=False, reflect=False):
        super().__init__(poly, seed, xor, reflect)


# Calculates 16-bit CRCs over a buffer of data.
class Crc16(Crc):
    # CRC-16-CCITT default polynomial
    DEFAULT_POLY = 0x1021
    # Seed for 16-bit CRC
    DEFAULT_SEED = 0xFFFF

    # Creates a 16 bit CRC algorithm instance. By default it is initialized to
    # use the CRC-16-CCITT algorithm.
    #
    # self.param poly [Integer] Polynomial to use when calculating the CRC
    # self.param seed [Integer] Seed value to start the calculation
    # self.param xor [Boolean] Whether to XOR the CRC result with 0xFFFF
    # self.param reflect [Boolean] Whether to bit reverse each byte of data before
    #   calculating the CRC
    def __init__(self, poly=DEFAULT_POLY, seed=DEFAULT_SEED, xor=False, reflect=False):
        super().__init__(poly, seed, xor, reflect)


# Calculates 32-bit CRCs over a buffer of data.
class Crc32(Crc):
    # CRC-32 default polynomial
    DEFAULT_POLY = 0x04C11DB7
    # Default Seed for 32-bit CRC
    DEFAULT_SEED = 0xFFFFFFFF

    # Creates a 32 bit CRC algorithm instance. By default it is initialized to
    # use the CRC-32 algorithm.
    #
    # self.param poly [Integer] Polynomial to use when calculating the CRC
    # self.param seed [Integer] Seed value to start the calculation
    # self.param xor [Boolean] Whether to XOR the CRC result with 0xFFFF
    # self.param reflect [Boolean] Whether to bit reverse each byte of data before
    #   calculating the CRC
    def __init__(self, poly=DEFAULT_POLY, seed=DEFAULT_SEED, xor=True, reflect=True):
        super().__init__(poly, seed, xor, reflect)


# Calculates 64-bit CRCs over a buffer of data.
class Crc64(Crc):
    # CRC-64-ECMA default polynomial
    DEFAULT_POLY = 0x42F0E1EBA9EA3693
    # Default Seed for 64-bit CRC
    DEFAULT_SEED = 0xFFFFFFFFFFFFFFFF

    # Creates a 64 bit CRC algorithm instance. By default it is initialized to
    # use the algorithm.
    #
    # self.param poly [Integer] Polynomial to use when calculating the CRC
    # self.param seed [Integer] Seed value to start the calculation
    # self.param xor [Boolean] Whether to XOR the CRC result with 0xFFFF
    # self.param reflect [Boolean] Whether to bit reverse each byte of data before
    #   calculating the CRC
    def __init__(self, poly=DEFAULT_POLY, seed=DEFAULT_SEED, xor=True, reflect=True):
        super().__init__(poly, seed, xor, reflect)



# Copyright 2024 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

# You can quickly setup an unauthenticated MQTT server in Docker with
# docker run -it -p 1883:1883 eclipse-mosquitto:2.0.15 mosquitto -c /mosquitto-no-auth.conf
# You can also test against encrypted and authenticated servers at https://test.mosquitto.org/

from time import sleep
import queue
import tempfile
from openc3.interfaces.interface import Interface
from openc3.system.system import System
from openc3.utilities.logger import Logger
from openc3.config.config_parser import ConfigParser

# See https://eclipse.dev/paho/files/paho.mqtt.python/html/client.html
import paho.mqtt.client as mqtt


# Base class for interfaces that send and receive messages over MQTT
class MqttInterface(Interface):
    # @param hostname [String] MQTT server to connect to
    # @param port [Integer] MQTT port
    def __init__(self, hostname, port=1883, ssl=False):
        super().__init__()
        self.hostname = hostname
        self.port = int(port)
        self.ssl = ConfigParser.handle_true_false(ssl)
        self.ack_timeout = 5.0
        self.username = None
        self.password = None
        self.cert = None
        self.key = None
        self.ca_file = None
        self.keyfile_password = None

        self.read_topics = []
        self.write_topics = []
        self.pkt_queue = queue.Queue()

        # Build list of packets by topic
        self.read_packets_by_topic = {}
        for _, target_packets in System.telemetry.all().items():
            for _, packet in target_packets.items():
                topics = packet.meta.get("TOPIC")
                if not topics:
                    topics = packet.meta.get("TOPICS")
                if topics:
                    for topic in topics:
                        self.read_packets_by_topic[topic] = packet

    def connection_string(self):
        return f"{self.hostname}:{self.port} (ssl: {self.ssl})"

    # Connects the interface to its target(s)
    def connect(self):
        self.read_topics = []
        self.write_topics = []
        self.pkt_queue.empty()

        self.client = mqtt.Client(mqtt.CallbackAPIVersion.VERSION2)
        self.client.on_connect = self.on_connect
        self.client.on_message = self.on_message
        self.client.user_data_set(self.pkt_queue)  # passed to on_message

        if self.ssl:
            self.client.tls_set()
        if self.username and self.password:
            self.client.username_pw_set(self.username, self.password)
        # You still need the ca_file if you're using your own cert and key
        if self.cert and self.key and self.ca_file:
            if self.keyfile_password:
                self.client.tls_set(
                    ca_certs=self.ca_file.name,
                    certfile=self.cert.name,
                    keyfile=self.key.name,
                    keyfile_password=self.keyfile_password,
                )
            else:
                self.client.tls_set(ca_certs=self.ca_file.name, certfile=self.cert.name, keyfile=self.key.name)
        elif self.ca_file:
            self.client.tls_set(ca_certs=self.ca_file.name)

        self.client.loop_start()
        # Connect doesn't fully establish the connection, it just sends the CONNECT packet
        # When the client loop receives an ONNACK packet from the broker in response to the CONNECT packet
        # it calls the on_connect callback and updates the client state to connected (is_connected() returns True)
        self.client.connect(self.hostname, self.port)
        i = 0
        while not self.client.is_connected() and i < (self.ack_timeout * 100):
            sleep(0.01)
            i += 1
        super().connect()

    def on_connect(self, client, userdata, flags, reason_code, properties):
        if reason_code.is_failure:
            Logger.error(f"MQTT failed to connect: {reason_code}")
        else:
            # we should always subscribe from on_connect callback to be sure
            # our subscribed is persisted across reconnections.
            for topic, _ in self.read_packets_by_topic.items():
                client.subscribe(topic)

    # @return [Boolean] Whether the MQTT client is connected
    def connected(self):
        if self.client:
            return self.client.is_connected()
        else:
            return False

    # Disconnects the interface from its target(s)
    def disconnect(self):
        self.pkt_queue.put(None)
        if self.client:
            self.client.disconnect()
            self.client = None
        super().disconnect()

    def read(self):
        packet = super().read()
        if packet is not None:
            topic = self.read_topics.pop(0)
            identified_packet = self.read_packets_by_topic.get(topic)
            if identified_packet:
                identified_packet = identified_packet.clone()
                identified_packet.buffer = packet.buffer
                packet = identified_packet
            packet.received_time = None
        return packet

    def write(self, packet):
        self.write_mutex.acquire()
        topics = packet.meta.get("TOPIC")
        if not topics:
            topics = packet.meta.get("TOPICS")
        if topics:
            for topic in topics:
                self.write_topics.append(topic)
                super().write(packet)
        else:
            raise RuntimeError(
                f"Command packet '{packet.target_name} {packet.packet_name}' requires a META TOPIC or TOPICS"
            )
        self.write_mutex.release()

    def on_message(self, client, userdata, message):
        # userdata is set via user_data_set
        self.read_topics.append(message.topic)
        userdata.put(message.payload)

    def read_interface(self):
        data = self.pkt_queue.get(block=True)
        extra = None
        if data is not None:
            self.read_interface_base(data, extra)
        return (data, extra)

    # Writes to the client
    # @param data [String] Raw packet data
    def write_interface(self, data, extra=None):
        self.write_interface_base(data, extra)
        try:
            topic = self.write_topics.pop(0)
        except IndexError:
            raise RuntimeError(f"write_interface called with no topics: {self.write_topics}")
        info = self.client.publish(topic, data)
        # This more closely matches the ruby implementation
        info.wait_for_publish(timeout=self.ack_timeout)

    # Supported Options
    # USERNAME - Username for Mqtt Server
    # PASSWORD - Password for Mqtt Server
    # CERT - Public Key for Client Cert Auth
    # KEY - Private Key for Client Cert Auth
    # CA_FILE - Certificate Authority for Client Cert Auth
    # KEYFILE_PASSWORD - Password for encrypted cert and key files
    # (see Interface#set_option)
    def set_option(self, option_name, option_values):
        super().set_option(option_name, option_values)
        match option_name.upper():
            case "ACK_TIMEOUT":
                self.ack_timeout = float(option_values[0])
            case "USERNAME":
                self.username = option_values[0]
            case "PASSWORD":
                self.password = option_values[0]
            case "CERT":
                # CERT must be given as a file
                self.cert = tempfile.NamedTemporaryFile(mode="w+", delete=False)
                self.cert.write(option_values[0])
                self.cert.close()
            case "KEY":
                # KEY must be given as a file
                self.key = tempfile.NamedTemporaryFile(mode="w+", delete=False)
                self.key.write(option_values[0])
                self.key.close()
            case "CA_FILE":
                # CA_FILE must be given as a file
                self.ca_file = tempfile.NamedTemporaryFile(mode="w+", delete=False)
                self.ca_file.write(option_values[0])
                self.ca_file.close()
            case "KEYFILE_PASSWORD":
                self.keyfile_password = option_values[0]



# Copyright 2025 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

from openc3.config.config_parser import ConfigParser
from openc3.interfaces.protocols.burst_protocol import BurstProtocol
from openc3.accessors.binary_accessor import BinaryAccessor


# Protocol which delineates packets using a length field at a fixed
# location in each packet.
class LengthProtocol(BurstProtocol):
    # self.param length_bit_offset [Integer] The bit offset of the length field
    # self.param length_bit_size [Integer] The size in bits of the length field
    # self.param length_value_offset [Integer] The offset to apply to the length
    #   value once it has been read from the packet. The value in the length
    #   field itself plus the length value offset MUST equal the total bytes
    #   including any discarded bytes.
    #   For example= if your length field really means "length - 1" this value should be 1.:
    # self.param length_bytes_per_count [Integer] The number of bytes per each
    #   length field 'count'. This is used if the units of the length field is:
    #   something other than bytes, for example words.
    # self.param length_endianness [String] The endianness of the length field.
    #   Must be either BIG_ENDIAN or LITTLE_ENDIAN.
    # self.param discard_leading_bytes (see BurstProtocol#initialize)
    # self.param sync_pattern (see BurstProtocol#initialize)
    # self.param max_length [Integer] The maximum allowed value of the length field
    # self.param fill_length_and_sync_pattern [Boolean] Fill the length field and sync
    #    pattern when writing packets
    # self.param allow_empty_data [True/False/None] See Protocol#initialize
    def __init__(
        self,
        length_bit_offset=0,
        length_bit_size=16,
        length_value_offset=0,
        length_bytes_per_count=1,
        length_endianness="BIG_ENDIAN",
        discard_leading_bytes=0,
        sync_pattern=None,
        max_length=None,
        fill_length_and_sync_pattern=False,
        allow_empty_data=None,
    ):
        super().__init__(
            discard_leading_bytes,
            sync_pattern,
            fill_length_and_sync_pattern,
            allow_empty_data,
        )

        # Save length field attributes
        self.length_bit_offset = int(length_bit_offset)
        self.length_bit_size = int(length_bit_size)
        self.length_value_offset = int(length_value_offset)
        self.length_bytes_per_count = int(length_bytes_per_count)

        # Save endianness
        if str(length_endianness).upper() == "LITTLE_ENDIAN":
            self.length_endianness = "LITTLE_ENDIAN"
        else:
            self.length_endianness = "BIG_ENDIAN"

        # Derive number of bytes required to contain entire length field
        if self.length_endianness == "BIG_ENDIAN" or ((self.length_bit_offset % 8) == 0):
            length_bits_needed = self.length_bit_offset + self.length_bit_size
            if (length_bits_needed % 8) != 0:
                length_bits_needed += 8
            self.length_bytes_needed = ((length_bits_needed - 1) / 8) + 1
        else:
            self.length_bytes_needed = (length_bit_offset / 8) + 1

        # Save max length setting
        self.max_length = ConfigParser.handle_none(max_length)
        if self.max_length:
            self.max_length = int(self.max_length)

    # Called to perform modifications on a command packet before it is send
    #
    # self.param packet [Packet] Original packet
    # self.return [Packet] Potentially modified packet
    def write_packet(self, packet):
        if self.fill_fields:
            # If the start of the length field is past what we discard, then the
            # length field is inside the packet
            if self.length_bit_offset >= (self.discard_leading_bytes * 8):
                length = self.calculate_length(len(packet.buffer_no_copy()) + self.discard_leading_bytes)
                # Subtract off the discarded bytes since they haven't been added yet
                # Adding bytes happens in the write_data method
                offset = self.length_bit_offset - (self.discard_leading_bytes * 8)
                # Directly write the packet buffer and fill in the length
                BinaryAccessor.write(
                    length,
                    offset,
                    self.length_bit_size,
                    "UINT",
                    packet.buffer_no_copy(),
                    self.length_endianness,
                    "ERROR",
                )
        return super().write_packet(packet)  # Allow burst_protocol to set the sync if needed:

    # Called to perform modifications on write data before making it into a packet
    #
    # self.param data [String] Raw packet data
    # self.return [String] Potentially modified packet data
    def write_data(self, data, extra=None):
        data, extra = super().write_data(data, extra)
        if self.fill_fields:
            # If the start of the length field is before what we discard, then the
            # length field is outside the packet
            if self.length_bit_offset < (self.discard_leading_bytes * 8):
                BinaryAccessor.write(
                    self.calculate_length(len(data)),
                    self.length_bit_offset,
                    self.length_bit_size,
                    "UINT",
                    data,
                    self.length_endianness,
                    "ERROR",
                )
        return (data, extra)

    def calculate_length(self, buffer_length):
        length = int(buffer_length / self.length_bytes_per_count) - self.length_value_offset
        if self.max_length and length > self.max_length:
            raise ValueError(f"Calculated length {length} larger than max_length {self.max_length}")
        return length

    def reduce_to_single_packet(self):
        # Make sure we have at least enough data to reach the length field
        if len(self.data) < self.length_bytes_needed:
            return ("STOP", self.extra)

        # Determine the packet's length
        length = BinaryAccessor.read(
            self.length_bit_offset,
            self.length_bit_size,
            "UINT",
            self.data,
            self.length_endianness,
        )
        if self.max_length and length > self.max_length:
            raise ValueError(f"Length value received larger than max_length= {length} > {self.max_length}")

        packet_length = (length * self.length_bytes_per_count) + self.length_value_offset
        # Ensure the calculated packet length is long enough to support the location of the length field
        # without overlap into the next packet
        if (packet_length * 8) < (self.length_bit_offset + self.length_bit_size):
            raise ValueError(
                f"Calculated packet length of {packet_length * 8} bits < (offset={self.length_bit_offset} + size={self.length_bit_size})"
            )

        # Make sure we have enough data for the packet
        if len(self.data) < packet_length:
            return ("STOP", self.extra)

        # Reduce to packet data and setup current_data for next packet
        packet_data = self.data[0:packet_length]
        self.data = self.data[packet_length:]

        return (packet_data, self.extra)



# Copyright 2024 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

# This file implements a class to handle command validation


# This class defines methods which are called when a command is sent.
# This class must be subclassed and the pre_check or
# post_check methods implemented. Do NOT use this class directly.
class CommandValidator:
    def __init__(self, command=None):
        self.command = command
        self.args = []

    def pre_check(self, command):
        # Return True to indicate Success, False to indicate Failure,
        # and None to indicate Unknown. The second value is the optional message.
        return [True, None]

    def post_check(self, command):
        # Return True to indicate Success, False to indicate Failure,
        # and None to indicate Unknown. The second value is the optional message.
        return [True, None]

    def args(self):
        return self.args



# Copyright 2024 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

from openc3.conversions.object_read_conversion import ObjectReadConversion


class ObjectWriteConversion(ObjectReadConversion):
    # Perform the conversion on the value.
    #
    # @param value [Object] Hash of packet key/value pairs
    # @param packet [Packet] Unused
    # @param buffer [String] The packet buffer
    # @return Raw BLOCK data
    def call(self, value, _packet, _buffer):
        fill_packet = self.lookup_packet()
        fill_packet.restore_defaults()
        for key, write_value in value.items():
            fill_packet.write(key, write_value)
        return fill_packet.buffer

    # @return [String] Config fragment for this conversion
    def to_config(self, read_or_write):
        return f"{read_or_write}_CONVERSION openc3/conversions/object_write_conversion.py {self.cmd_or_tlm if self.cmd_or_tlm else 'None'} {self.target_name} {self.packet_name}\n"



# Copyright 2025 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.
#
# A portion of this file was funded by Blue Origin Enterprises, L.P.
# See https://github.com/OpenC3/cosmos/pull/1953

import re
from .accessor import Accessor
from openc3.top_level import get_class_from_module
from openc3.utilities.string import class_name_to_filename, filename_to_module, filename_to_class_name


class HttpAccessor(Accessor):
    def __init__(self, packet, body_accessor="openc3/accessors/form_accessor.py", *body_accessor_args):
        super().__init__(packet)
        self.args.append(body_accessor)
        for arg in body_accessor_args:
            self.args.append(arg)
        try:
            klass = get_class_from_module(
                filename_to_module(body_accessor),
                filename_to_class_name(body_accessor),
            )
        # Fall back to the deprecated behavior of passing the ClassName
        except ModuleNotFoundError:
            filename = class_name_to_filename(body_accessor)
            klass = get_class_from_module(f"openc3.accessors.{filename}", body_accessor)

        self.body_accessor = klass(packet, *body_accessor_args)

    def read_item(self, item, buffer):
        if item.name in [
            "HTTP_STATUS",
            "HTTP_PATH",
            "HTTP_METHOD",
            "HTTP_PACKET",
            "HTTP_ERROR_PACKET",
        ]:
            if not self.packet.extra:
                return None
            return self.packet.extra[item.name]

        item_name = item.name
        if re.match(r"^HTTP_QUERY_", item_name):
            if not self.packet.extra:
                return None
            if re.match(r"^HTTP_QUERY_", item.key):
                query_name = item_name[11:].lower()
            else:
                query_name = item.key
            queries = self.packet.extra.get("HTTP_QUERIES")
            if queries:
                return queries[query_name]
            else:
                return None

        if re.match(r"^HTTP_HEADER_", item_name):
            if not self.packet.extra:
                return None
            if re.match(r"^HTTP_HEADER_", item.key):
                header_name = item_name[12:].lower()
            else:
                header_name = item.key
            headers = self.packet.extra.get("HTTP_HEADERS")
            if headers:
                return headers[header_name]
            else:
                return None

        return self.body_accessor.read_item(item, buffer)

    def write_item(self, item, value, buffer):
        if item.name == "HTTP_STATUS":
            self.packet.extra = self.packet.extra or {}
            self.packet.extra[item.name] = int(value)
            return self.packet.extra[item.name]

        if item.name in [
            "HTTP_PATH",
            "HTTP_METHOD",
            "HTTP_PACKET",
            "HTTP_ERROR_PACKET",
        ]:
            self.packet.extra = self.packet.extra or {}
            value = str(value)
            if item.name == "HTTP_METHOD":
                value = value.lower()
            elif item.name in ["HTTP_PACKET", "HTTP_ERROR_PACKET"]:
                value = value.upper()
            self.packet.extra[item.name] = value
            return self.packet.extra[item.name]

        item_name = item.name
        if re.match(r"^HTTP_QUERY_", item_name):
            self.packet.extra = self.packet.extra or {}
            if re.match(r"^HTTP_QUERY_", item.key):
                query_name = item_name[11:].lower()
            else:
                query_name = item.key
            self.packet.extra["HTTP_QUERIES"] = self.packet.extra.get("HTTP_QUERIES", {})
            queries = self.packet.extra["HTTP_QUERIES"]
            queries[query_name] = str(value)
            return queries[query_name]

        if re.match(r"^HTTP_HEADER_", item_name):
            self.packet.extra = self.packet.extra or {}
            if re.match(r"^HTTP_HEADER_", item.key):
                header_name = item_name[12:].lower()
            else:
                header_name = item.key
            self.packet.extra["HTTP_HEADERS"] = self.packet.extra.get("HTTP_HEADERS", {})
            headers = self.packet.extra["HTTP_HEADERS"]
            headers[header_name] = str(value)
            return headers[header_name]

        self.body_accessor.write_item(item, value, buffer)
        return value

    def read_items(self, items, buffer):
        result = {}
        body_items = []
        for item in items:
            if item.name[0:4] == "HTTP_":
                result[item.name] = self.read_item(item, buffer)
            else:
                body_items.append(item)
        body_result = self.body_accessor.read_items(body_items, buffer)
        return result | body_result  # Merge Body accessor read items with HTTP_ items

    def write_items(self, items, values, buffer):
        body_items = []
        for item, index in enumerate(items):
            if item.name[0:4] == "HTTP_":
                self.write_item(item, values[index], buffer)
            else:
                body_items.append(item)
        self.body_accessor.write_items(body_items, values, buffer)
        return values

    # If this is set it will enforce that buffer data is encoded
    # in a specific encoding
    def enforce_encoding(self):
        return self.body_accessor.enforce_encoding()

    # This affects whether the Packet class enforces the buffer
    # length at all.  Set to false to remove any correlation between
    # buffer length and defined sizes of items in COSMOS
    def enforce_length(self):
        return self.body_accessor.enforce_length()

    # This sets the short_buffer_allowed flag in the Packet class
    # which allows packets that have a buffer shorter than the defined size.
    # Note that the buffer is still resized to the defined length
    def enforce_short_buffer_allowed(self):
        return self.body_accessor.enforce_short_buffer_allowed()

    # If this is true it will enforce that COSMOS DERIVED items must have a
    # write_conversion to be written
    def enforce_derived_write_conversion(self, item):
        if item.name in [
            "HTTP_STATUS",
            "HTTP_PATH",
            "HTTP_METHOD",
            "HTTP_PACKET",
            "HTTP_ERROR_PACKET",
        ]:
            return False
        if re.match(r"^HTTP_QUERY_", item.name) or re.match(r"^HTTP_HEADER_", item.name):
            return False

        return self.body_accessor.enforce_derived_write_conversion(item)



# Copyright 2024 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.


class Conversion:
    """Performs a general conversion via the implementation of the call method"""

    # self.return [Symbol] The converted data type. Must be one of
    #   {OpenC3::StructureItem#data_type}
    # attr_reader :converted_type
    # # self.return [Integer] The size in bits of the converted value
    # attr_reader :converted_bit_size
    # # self.return [Integer] The size in bits of the converted array value
    # attr_reader :converted_array_size
    # attr_reader :params

    # Create a new conversion
    def __init__(self):
        self.converted_type = None
        self.converted_bit_size = None
        self.converted_array_size = None
        self.params = None

    # Perform the conversion on the value.
    #
    # self.param value [Object] The value to convert
    # self.param packet [Packet] The packet which contains the value. This can
    #   be useful to reach into the packet and use other values in the
    #   conversion.
    # self.param buffer [String] The packet buffer
    # self.return The converted value
    def call(self, value, packet, buffer):
        raise RuntimeError("call method must be defined by subclass")

    # self.return [String] The conversion class
    def __str__(self):
        return self.__class__.__name__

    # self.param read_or_write [String] Either 'READ' or 'WRITE'
    # self.return [String] Config fragment for this conversion
    def to_config(self, read_or_write):
        return f"    {read_or_write}_CONVERSION {self.__class__.__name__}\n"

    def as_json(self):
        result = {}
        result["class"] = self.__class__.__name__
        if self.converted_type is not None:
            result["converted_type"] = self.converted_type
        if self.converted_bit_size is not None:
            result["converted_bit_size"] = self.converted_bit_size
        if self.converted_array_size is not None:
            result["converted_array_size"] = self.converted_array_size
        if self.params is not None:
            result["params"] = self.params
        return result



# Copyright 2023 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

from typing import Any

from openc3.utilities.store import Store
from openc3.utilities.local_mode import LocalMode
from openc3.environment import OPENC3_SCOPE


class ToolConfigModel:
    @classmethod
    def config_tool_names(cls, scope: str = OPENC3_SCOPE):
        cursor, keys = Store.scan(match=f"{scope}__config__*", count=100, _type="HASH")
        result = [key.decode().split("__")[2] for key in keys]
        result.sort()
        return result

    @classmethod
    def list_configs(cls, tool: str, scope: str = OPENC3_SCOPE):
        keys = Store.hkeys(f"{scope}__config__{tool}")
        return [key.decode() for key in keys]

    @classmethod
    def load_config(cls, tool: str, name: str, scope: str = OPENC3_SCOPE):
        return Store.hget(f"{scope}__config__{tool}", name).decode()

    @classmethod
    def save_config(
        cls,
        tool: str,
        name: str,
        data: Any,
        local_mode: bool = True,
        scope: str = OPENC3_SCOPE,
    ):
        Store.hset(f"{scope}__config__{tool}", name, data)
        if local_mode:
            LocalMode.save_tool_config(scope, tool, name, data)

    @classmethod
    def delete_config(cls, tool: str, name: str, local_mode: bool = True, scope: str = OPENC3_SCOPE):
        Store.hdel(f"{scope}__config__{tool}", name)
        if local_mode:
            LocalMode.delete_tool_config(scope, tool, name)



# Copyright 2023 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

from openc3.packets.packet_item import PacketItem
from openc3.utilities.logger import Logger
from openc3.utilities.extract import convert_to_value


class StateParser:
    COMMAND = "Command"
    TELEMETRY = "Telemetry"

    # self.param parser [ConfigParser] Configuration parser
    # self.param packet [Packet] The current packet
    # self.param cmd_or_tlm [String] Whether this is a command or telemetry packet
    # self.param item [PacketItem] The packet item to create states on
    # self.param warnings [Array<String>] Array of string warnings which will be
    #   appended with any warnings found case parsing the limits:
    @classmethod
    def parse(cls, parser, packet, cmd_or_tlm, item, warnings):
        if item.limits.values:
            raise parser.error("Items with LIMITS can't define STATE")
        if item.units:
            raise parser.error("Items with UNITS can't define STATE")

        state_parser = StateParser(parser)
        state_parser.verify_parameters(cmd_or_tlm)
        state_parser.create_state(packet, cmd_or_tlm, item, warnings)

    # self.param parser [ConfigParser] Configuration parser
    def __init__(self, parser):
        self.parser = parser

    # self.param cmd_or_tlm [String] Whether this is a command or telemetry packet
    def verify_parameters(self, cmd_or_tlm):
        self.usage = "STATE <STATE NAME> <STATE VALUE> "
        if cmd_or_tlm == StateParser.COMMAND:
            self.usage += "<HAZARDOUS / DISABLE_MESSAGES (Optional)> <Hazardous Description (Optional)>"
            self.parser.verify_num_parameters(2, 4, self.usage)
        else:
            self.usage += "<COLOR: GREEN/YELLOW/RED (Optional)>"
            self.parser.verify_num_parameters(2, 3, self.usage)

    # self.param packet [Packet] The current packet
    # self.param cmd_or_tlm [String] Whether this is a command or telemetry packet
    # self.param item [PacketItem] The packet item to create states on
    # self.param warnings [Array<String>] Array of string warnings which will be
    #   appended with any warnings found case parsing the limits:
    def create_state(self, packet, cmd_or_tlm, item, warnings):
        if item.states is None:
            item.states = {}

        state_name = self._get_state_name()
        self._check_for_duplicate_states(item, warnings)
        # Get the states, set one, then reassign for the setter to work
        states = item.states
        states[state_name] = self._get_state_value(item.data_type)
        item.states = states
        self._parse_additional_parameters(packet, cmd_or_tlm, item)

    def _get_state_name(self):
        return self.parser.parameters[0].upper()

    def _get_state_value(self, data_type):
        if data_type == "STRING" or data_type == "BLOCK":
            return self.parser.parameters[1]
        else:
            return convert_to_value(self.parser.parameters[1])

    def _check_for_duplicate_states(self, item, warnings):
        if item.states.get(self._get_state_name()) is not None:
            msg = f"Duplicate state defined on line {self.parser.line_number}: {self.parser.line}"
            Logger.warn(msg)
            warnings.append(msg)

    def _parse_additional_parameters(self, packet, cmd_or_tlm, item):
        if not len(self.parser.parameters) > 2:
            return

        if cmd_or_tlm == StateParser.COMMAND:
            self._get_hazardous_or_disable_messages(item)
        else:
            self._get_state_colors(item)
            packet.update_limits_items_cache(item)

    def _get_hazardous_or_disable_messages(self, item):
        match self.parser.parameters[2].upper():
            case "HAZARDOUS":
                if item.hazardous is None:
                    item.hazardous = {}
                if len(self.parser.parameters) >= 4:
                    item.hazardous[self._get_state_name()] = self.parser.parameters[3]
                else:
                    item.hazardous[self._get_state_name()] = ""
            case "DISABLE_MESSAGES":
                if item.messages_disabled is None:
                    item.messages_disabled = {}
                item.messages_disabled[self._get_state_name()] = True
            case _:
                raise self.parser.error(
                    "HAZARDOUS or DISABLE_MESSAGES expected as third parameter for this line.",
                    self.usage,
                )

    def _get_state_colors(self, item):
        color = self.parser.parameters[2].upper()
        if color not in PacketItem.STATE_COLORS:
            raise self.parser.error(
                f"Invalid state color {color}. Must be one of {' '.join(PacketItem.STATE_COLORS)}.",
                self.usage,
            )

        item.limits.enabled = True
        if item.state_colors is None:
            item.state_colors = {}
        item.state_colors[self._get_state_name()] = color



# Copyright 2023 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

from openc3.config.config_parser import ConfigParser
from openc3.interfaces.protocols.protocol import Protocol
from openc3.accessors.binary_accessor import BinaryAccessor
from openc3.utilities.logger import Logger
from openc3.utilities.crc import Crc8, Crc16, Crc32, Crc64


# Creates a CRC on write and verifies a CRC on read
class CrcProtocol(Protocol):
    ERROR = "ERROR"  # on CRC mismatch
    DISCONNECT = "DISCONNECT"  # on CRC mismatch

    # self.param write_item_name [String/None] Item to fill with calculated CRC value for outgoing packets (None = don't fill)
    # self.param strip_crc [Boolean] Whether or not to remove the CRC from incoming packets
    # self.param bad_strategy [ERROR/DISCONNECT] How to handle CRC errors on incoming packets.  ERROR = Just log the error, DISCONNECT = Disconnect interface
    # self.param bit_offset [Integer] Bit offset of the CRC in the data.  Can be negative to indicate distance from end of packet
    # self.param bit_size [Integer] Bit size of the CRC - Must be 16, 32, or 64
    # self.param endianness [BIG_ENDIAN/LITTLE_ENDIAN] Endianness of the CRC
    # self.param poly [Integer] Polynomial to use when calculating the CRC
    # self.param seed [Integer] Seed value to start the calculation
    # self.param xor [Boolean] Whether to XOR the CRC result with 0xFFFF
    # self.param reflect [Boolean] Whether to bit reverse each byte of data before calculating the CRC
    # self.param allow_empty_data [True/False/None] See Protocol#initialize
    def __init__(
        self,
        write_item_name=None,
        strip_crc=False,
        bad_strategy="ERROR",
        bit_offset=-32,
        bit_size=32,
        endianness="BIG_ENDIAN",
        poly=None,
        seed=None,
        xor=None,
        reflect=None,
        allow_empty_data=None,
    ):
        super().__init__(allow_empty_data)
        self.write_item_name = ConfigParser.handle_none(write_item_name)
        self.strip_crc = ConfigParser.handle_true_false(strip_crc)
        if self.strip_crc is not True and self.strip_crc is not False:
            raise ValueError(f"Invalid strip CRC of '{strip_crc}'. Must be TRUE or FALSE.")

        match bad_strategy:
            case CrcProtocol.ERROR | CrcProtocol.DISCONNECT:
                self.bad_strategy = bad_strategy
            case _:
                raise ValueError(f"Invalid bad CRC strategy of {bad_strategy}. Must be ERROR or DISCONNECT.")

        match str(endianness).upper():
            case "BIG_ENDIAN":
                self.endianness = "BIG_ENDIAN"  # Convert to symbol for use in BinaryAccessor.write
            case "LITTLE_ENDIAN":
                self.endianness = "LITTLE_ENDIAN"  # Convert to symbol for use in BinaryAccessor.write
            case _:
                raise ValueError("Invalid endianness '{endianness}'. Must be BIG_ENDIAN or LITTLE_ENDIAN.")

        try:
            self.bit_offset = int(bit_offset)
        except TypeError:
            raise ValueError(f"Invalid bit offset of {bit_offset}. Must be a number.")
        if self.bit_offset % 8 != 0:
            raise ValueError(f"Invalid bit offset of {bit_offset}. Must be divisible by 8.")

        poly = ConfigParser.handle_none(poly)
        try:
            if isinstance(poly, str):
                poly = int(poly, 0)
        except (ValueError, TypeError):
            raise ValueError(f"Invalid polynomial of {poly}. Must be a number.")

        seed = ConfigParser.handle_none(seed)
        try:
            if isinstance(seed, str):
                seed = int(seed, 0)
        except ValueError:
            raise ValueError(f"Invalid seed of {seed}. Must be a number.")

        xor = ConfigParser.handle_true_false_none(xor)
        if xor is not None and xor is not True and xor is not False:
            raise ValueError(f"Invalid XOR value of '{xor}'. Must be TRUE or FALSE.")

        reflect = ConfigParser.handle_true_false_none(reflect)
        if reflect is not None and reflect is not True and reflect is not False:
            raise ValueError(f"Invalid reflect value of '{reflect}'. Must be TRUE or FALSE.")

        # Built the CRC arguments array. All subsequent arguments are dependent
        # on the previous ones so we build it up incrementally.
        args = []
        if poly is not None:
            args.append(poly)
            if seed is not None:
                args.append(seed)
                if xor is not None:
                    args.append(xor)
                    if reflect is not None:
                        args.append(reflect)

        try:
            self.bit_size = int(bit_size)
        except TypeError:
            raise ValueError(f"Invalid bit size of {bit_size}. Must be a number.")
        endianness = "<"  # LITTLE_ENDIAN
        if self.endianness == "BIG_ENDIAN":
            endianness = ">"
        match self.bit_size:
            case 8:
                self.pack = f"{endianness}B"
                if len(args) == 0:
                    self.crc = Crc8()
                else:
                    self.crc = Crc8(*args)
            case 16:
                self.pack = f"{endianness}H"
                if len(args) == 0:
                    self.crc = Crc16()
                else:
                    self.crc = Crc16(*args)
            case 32:
                self.pack = f"{endianness}I"
                if len(args) == 0:
                    self.crc = Crc32()
                else:
                    self.crc = Crc32(*args)
            case 64:
                self.pack = f"{endianness}Q"
                if len(args) == 0:
                    self.crc = Crc64()
                else:
                    self.crc = Crc64(*args)
            case _:
                raise ValueError(f"Invalid bit size of {bit_size}. Must be 16, 32, or 64.")

    def read_data(self, data, extra):
        if len(data) <= 0:
            return super().read_data(data, extra)

        crc = BinaryAccessor.read(self.bit_offset, self.bit_size, "UINT", data, self.endianness)
        calculated_crc = self.crc.calc(data[0 : int(self.bit_offset / 8)])
        if calculated_crc != crc:
            interface = ""
            if self.interface:
                interface = self.interface.name
            Logger.error(f"{interface}: Invalid CRC detected! Calculated {hex(calculated_crc)} vs found {hex(crc)}.")
            if self.bad_strategy == CrcProtocol.DISCONNECT:
                return ("DISCONNECT", extra)
        if self.strip_crc:
            new_data = data[:]
            new_data = new_data[0 : int(self.bit_offset / 8)]
            end_range = int((self.bit_offset + self.bit_size) / 8)
            if end_range != 0:
                new_data += data[end_range:]
            return (new_data, extra)
        return (data, extra)

    def write_packet(self, packet):
        if self.write_item_name:
            end_range = int(packet.get_item(self.write_item_name).bit_offset / 8)
            crc = self.crc.calc(packet.buffer_no_copy()[0:end_range])
            packet.write(self.write_item_name, crc)
        return packet

    def write_data(self, data, extra):
        if not self.write_item_name:
            if self.bit_size == 64:
                crc = self.crc.calc(data)
                data += b"\x00" * 8
                BinaryAccessor.write((crc >> 32), -64, 32, "UINT", data, self.endianness, "ERROR")
                BinaryAccessor.write((crc & 0xFFFFFFFF), -32, 32, "UINT", data, self.endianness, "ERROR")
            else:
                crc = self.crc.calc(data)
                data += b"\x00" * int(self.bit_size / 8)
                BinaryAccessor.write(
                    crc,
                    -self.bit_size,
                    self.bit_size,
                    "UINT",
                    data,
                    self.endianness,
                    "ERROR",
                )
        return (data, extra)



# Copyright 2023 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

# Adds IO streams and then defers to the streams when using any of the Python
# output methods such as print

import sys


class IoMultiplexer:
    STDOUT = sys.stdout
    STDERR = sys.stderr

    # Create the empty stream array
    def __init__(self):
        self.streams = []

    # Delegate all unknown methods to all streams
    def __getattr__(self, func):
        def method(*args, **kwargs):
            first = True
            result = None
            for stream in self.streams:
                if first:
                    result = getattr(stream, func)(*args, **kwargs)
                    if result == stream:
                        result = self
                    first = False
                else:
                    getattr(stream, func)(*args, **kwargs)
            return result

        return method

    # Removes STDOUT and STDERR from the array of streams
    def remove_default_io(self):
        if self.STDOUT in self.streams:
            self.streams.remove(self.STDOUT)
        if self.STDERR in self.streams:
            self.streams.remove(self.STDERR)

    # @param stream [IO] The stream to add
    def add_stream(self, stream):
        if stream not in self.streams:
            self.streams.append(stream)

    # @param stream [IO] The stream to remove
    def remove_stream(self, stream):
        if stream in self.streams:
            self.streams.remove(stream)



# Copyright 2025 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

from openc3.packets.packet import Packet
from openc3.tools.table_manager.table_item import TableItem

class Table(Packet):
    """
    Table extends Packet by adding more attributes relative to
    displaying binary data in a gui.
    """

    # Define the target for tables as 'TABLE' since there is no target
    TARGET = 'TABLE'

    def __init__(self, name, endianness, type_param, description, filename):
        """
        Constructor for a TableDefinition

        Args:
            name: Name of the table
            endianness: Byte order of the table (BIG_ENDIAN or LITTLE_ENDIAN)
            type_param: Either :KEY_VALUE or :ROW_COLUMN
            description: Description of the table
            filename: File which contains the table definition
        """
        super().__init__(self.TARGET, name, endianness, description, b'', item_class=TableItem)
        # ONE_DIMENSIONAL and TWO_DIMENSIONAL are deprecated so translate
        if type_param == "ONE_DIMENSIONAL":
            type_param = "KEY_VALUE"
        if type_param == "TWO_DIMENSIONAL":
            type_param = "ROW_COLUMN"

        if type_param != "KEY_VALUE" and type_param != "ROW_COLUMN":
            raise ValueError(f"Invalid type '{type_param}' for table '{name}'. Must be KEY_VALUE or ROW_COLUMN")

        self.type = type_param
        self.filename = filename
        self._num_rows = 0
        self._num_columns = 1 if self.type == "KEY_VALUE" else 0

    @property
    def table_name(self):
        """Table name"""
        return self.packet_name
        
    @property
    def endianness(self):
        """Get the table's endianness"""
        return self.default_endianness
        
    @property
    def length(self):
        """Returns the length of the table in bytes"""
        return self.defined_length

    @property
    def num_columns(self):
        """Number of columns in the table"""
        return self._num_columns

    @num_columns.setter
    def num_columns(self, value):
        """Set number of columns in the table"""
        self._num_columns = value

    @property
    def num_rows(self):
        """
        Get the number of rows in the table

        Returns:
            Number of rows in the table
        """
        if self.type == "KEY_VALUE":
            return len([item for item in self.sorted_items if not item.hidden])
        else:  # ROW_COLUMN
            return self._num_rows

    @num_rows.setter
    def num_rows(self, num_rows):
        """
        Set the number of rows in a ROW_COLUMN table

        Args:
            num_rows: Number of rows

        Raises:
            ValueError if table type is KEY_VALUE
        """
        if self.type == "KEY_VALUE":
            raise ValueError("Rows are fixed in a KEY_VALUE table")
        elif self.type == "ROW_COLUMN":
            self._num_rows = num_rows


# Copyright 2023 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

from openc3.utilities.store import EphemeralStore


class TopicMeta(type):
    def __getattr__(cls, func):
        def method(*args, **kwargs):
            return getattr(EphemeralStore.instance(), func)(*args, **kwargs)

        return method


class Topic(metaclass=TopicMeta):
    @classmethod
    def clear_topics(cls, topics, maxlen=0):
        for topic in topics:
            EphemeralStore.xtrim(topic, maxlen)

    @classmethod
    def topics(cls, key, scope):
        return sorted(set(list(EphemeralStore.scan_iter(match=f"{scope}__{key}__*", type="stream", count=100))))

    @classmethod
    def get_cnt(cls, topic):
        _, packet = EphemeralStore.get_newest_message(topic)
        if packet:
            return int(packet[b"received_count"])
        else:
            return 0



# Copyright 2025 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

from openc3.models.model import Model

# Note: This model is locked to the DEFAULT scope
class OfflineAccessModel(Model):
    PRIMARY_KEY = "DEFAULT__openc3__offline_access"

    # NOTE: The following three class methods are used by the ModelController
    # and are reimplemented to enable various Model class methods to work
    @classmethod
    def get(cls, name: str, scope: str):
        return super().get(OfflineAccessModel.PRIMARY_KEY, name=name)

    @classmethod
    def names(cls, scope: str):
        return super().names(OfflineAccessModel.PRIMARY_KEY)

    @classmethod
    def all(cls, scope: str):
        return super().all(OfflineAccessModel.PRIMARY_KEY)

    def __init__(
        self, name: str, offline_access_token: str = None, updated_at: float = None, scope: str = 'DEFAULT'
    ):
        super().__init__(OfflineAccessModel.PRIMARY_KEY, name=name, updated_at=updated_at, scope='DEFAULT')
        self.offline_access_token = offline_access_token

    # @return [Hash] JSON encoding of this model
    def as_json(self):
        return {
            "name": self.name,
            "updated_at": self.updated_at,
            "offline_access_token": self.offline_access_token,
            "scope": self.scope,
        }



# Copyright 2023 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

from openc3.system.system import System
from openc3.interfaces.protocols.protocol import Protocol


# Ignore a specific packet by not letting it through the protocol
class IgnorePacketProtocol(Protocol):
    # @param target_name [String] Target name
    # @param packet_name [String] Packet name
    def __init__(self, target_name, packet_name, allow_empty_data=None):
        super().__init__(allow_empty_data)
        System.telemetry.packet(target_name, packet_name)
        self.target_name = target_name
        self.packet_name = packet_name

    def read_packet(self, packet):
        # Need to make sure packet is identified and defined
        target_names = None
        if self.interface:
            target_names = self.interface.tlm_target_names
        identified_packet = System.telemetry.identify_and_define_packet(packet, target_names)
        if identified_packet:
            if identified_packet.target_name == self.target_name and identified_packet.packet_name == self.packet_name:
                return "STOP"
        return super().read_packet(packet)

    def write_packet(self, packet):
        if packet.target_name == self.target_name and packet.packet_name == self.packet_name:
            return "STOP"

        return super().write_packet(packet)
